# NIST-developed software is provided by NIST as a public service. You may use, copy and distribute copies of the software in any medium, provided that you keep intact this entire notice. You may improve, modify and create derivative works of the software or any portion of the software, and you may copy and distribute such modifications or works. Modified works should carry a notice stating that you changed the software and should note the date and nature of any such change. Please explicitly acknowledge the National Institute of Standards and Technology as the source of the software.
# NIST-developed software is expressly provided "AS IS." NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED, IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
# You are solely responsible for determining the appropriateness of using and distributing the software and you assume all risks associated with its use, including but not limited to the risks and costs of program errors, compliance with applicable laws, damage to or loss of data, programs or equipment, and the unavailability or interruption of operation. This software is not intended to be used in any situation where a failure could cause risk of injury or damage to property. The software developed by NIST employees is not subject to copyright protection within the United States.
import json
import logging
import os
from typing import Sequence, Any, Callable
import copy
import numpy as np
import math
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.tensorboard import SummaryWriter
from torchtext.data.iterator import Iterator as TextDataIterator
import torch.nn.utils.clip_grad as torch_clip_grad

from collections import OrderedDict
from transformers import get_linear_schedule_with_warmup
from trojai.modelgen import conlleval

from .default_optimizer import defaultdict
from .default_optimizer import split_val_clean_trig
from .default_optimizer import train_val_dataset_split
from .default_optimizer import DefaultOptimizer
from .datasets import CSVTextDataset
from .training_statistics import EpochStatistics, EpochTrainStatistics, EpochValidationStatistics
from .default_optimizer import _running_eval_acc, _save_nandata, _validate_soft_to_hard_args
from .config import DefaultOptimizerConfig
from .constants import VALID_OPTIMIZERS, MAX_EPOCHS

logger = logging.getLogger(__name__)


class Ner_Metrics:
    def __init__(self):
        self.epoch_stats = {}
        self.best_epoch_stats = {}

    def get_stats(self, counts):
        stats = OrderedDict()
        c = counts
        overall, by_type = conlleval.metrics(counts)

        stats['tokens_processed'] = c.token_counter
        stats['phrases'] = c.found_correct
        stats['found'] = c.found_guessed
        stats['correct'] = c.correct_chunk
        stats['accuracy'] = c.correct_tags / c.token_counter
        stats['precision'] = overall.prec
        stats['recall'] = overall.rec
        stats['f1'] = overall.fscore

        for i, m in sorted(by_type.items()):
            type_stats = OrderedDict()
            type_stats['accuracy'] = m.tp / (m.fn + m.tp)
            type_stats['precision'] = m.prec
            type_stats['recall'] = m.rec
            type_stats['f1'] = m.fscore
            type_stats['guessed'] = c.t_found_guessed[i]
            stats[i] = type_stats
        return stats

    def add_epoch_stats(self, epoch_num, clean_counts, trigger_counts):
        res = {}
        if clean_counts is not None:
            res['eval_clean'] = self.get_stats(clean_counts)
        if trigger_counts is not None:
            res['eval_triggered'] = self.get_stats(trigger_counts)
        self.epoch_stats[epoch_num] = res

    def add_best(self, epoch_num, test_counts, clean_counts, trigger_counts):
        if test_counts is not None:
            self.best_epoch_stats['test'] = self.get_stats(test_counts)
        if clean_counts is not None:
            self.best_epoch_stats['test_clean'] = self.get_stats(clean_counts)
        if trigger_counts is not None:
            self.best_epoch_stats['test_triggered'] = self.get_stats(trigger_counts)

        self.best_epoch_stats['epoch_num'] = epoch_num

    def write_per_epoch(self, filepath):
        with open(filepath, 'w') as fp:
            json.dump(self.epoch_stats, fp, indent=2)

    def write_best_epoch(self, filepath):
        with open(filepath, 'w') as fp:
            json.dump(self.best_epoch_stats, fp, indent=2)


def _running_eval_acc(y_hat: torch.Tensor, y_truth: torch.Tensor, label_mask,
                      n_total: defaultdict = None, n_correct: defaultdict = None,
                      soft_to_hard_fn: Callable = None,
                      soft_to_hard_fn_kwargs: dict = None):
    """
    Wrapper for computing accuracy in an on-line manner
    :param y_hat: the computed predictions, should be of shape (n_batches, num_output_neurons)
    :param y_truth: the actual y-values
    :param n_total: a defaultdict with keys representing labels, and values representing the # of times examples
        with that label have been seen.  Example: {0: 10, 1: 20, 2: 5, 3: 30}
    :param n_correct: a defaultdict with keys representing labels, and values representing the # of times examples
        with that label have been corrected.  Example: {0: 8, 1: 15, 2: 5, 3: 25}
    :param soft_to_hard_fn: A function handle which takes y_hat and produces a hard-decision
    :param soft_to_hard_fn_kwargs: kwargs to pass to soft_to_hard_fn
    :return: accuracy, updated n_total, updated n_correct
    """

    y_hat_preds = torch.argmax(y_hat, dim=2)

    if not n_total:
        n_total = defaultdict(int)

    if not n_correct:
        n_correct = defaultdict(int)

    for batch in range(label_mask.shape[0]):
        mask = label_mask.data[batch]
        preds = y_hat_preds.data[batch]
        labels = y_truth.data[batch]

        for i, m in enumerate(mask):
            if m == 1:
                n_total[labels[i].item()] += 1
                n_correct[labels[i].item()] += preds[i].item() == labels[i].item()

    acc = 0.
    weight = 1. / len(n_total.keys())
    # 1.0 - n_total[class] /  sum(n_total) = weight
    # Apply per class weight, sum to 1.0

    for k in n_total.keys():
        val = 0
        if n_total[k] > 0:
            val = float(n_correct[k]) / float(n_total[k])
        acc += val

    acc *= 100. * weight

    return acc, n_total, n_correct


class TorchTextOptimizerShiftingEmbedding(DefaultOptimizer):
    """
    An optimizer for training and testing LSTM models. Currently in a prototype state.
    """

    def __init__(self, tokenizer, id2label, optimizer_cfg: DefaultOptimizerConfig = None, ner_report_dirpath: str = "./", run_id: str = ''):
        # def __init__(self, tokenizer, embedding, cls_token_is_first, optimizer_cfg: DefaultOptimizerConfig = None):
        """
        Initializes the optimizer with an DefaultOptimizerConfig
        :param tokenizer: the tokenizer to apply to text input data
        :param embedding: the embedding to apply to tokenized data
        :param optimizer_cfg: the configuration used to initialize the TorchTextOptimizer
        """
        super().__init__(optimizer_cfg)

        # Setup tokenizer and embedding
        self.tokenizer = tokenizer
        self.id2label = id2label
        # TODO: pass in as parameters
        self.gradient_accumulation_steps = 1
        self.warmup_steps = 0

        if self.tokenizer.name_or_path == 'gpt2':
            self.pad_token = self.tokenizer.eos_token
            self.cls_token = self.tokenizer.eos_token
            self.sep_token = self.tokenizer.eos_token
        else:
            self.pad_token = self.tokenizer.pad_token
            self.cls_token = self.tokenizer.cls_token
            self.sep_token = self.tokenizer.sep_token

        if not os.path.isdir(ner_report_dirpath):
            try:
                os.makedirs(ner_report_dirpath)
            except IOError as e:
                logger.error(e)

        self.ner_report_dirpath = ner_report_dirpath
        self.run_id = run_id

        self.ner_metrics = Ner_Metrics()
        self.best_epoch = 0

        # self.embedding = embedding
        # self.embedding.to(self.device)
        # self.max_input_length = self.tokenizer.max_model_input_sizes[self.tokenizer.name_or_path]
        # self.cls_token_is_first = cls_token_is_first

        # if not hasattr(self.tokenizer, 'pad_token') or self.tokenizer.pad_token is None:
        #     self.tokenizer.pad_token = self.tokenizer.eos_token

    def __deepcopy__(self, memodict={}):
        optimizer_cfg_copy = copy.deepcopy(self.optimizer_cfg)
        # WARNING: this assumes that none of the derived attributes have been changed after construction!
        return TorchTextOptimizerShiftingEmbedding(DefaultOptimizerConfig(optimizer_cfg_copy.training_cfg,
                                                                          optimizer_cfg_copy.reporting_cfg))

    def _eval_acc(self, data_loader, model,
                  loss_fn: Callable = None):
        """
        Evaluates a model against a dataset encompassed by a data loader, which has
        an underlying torchtext dataset.  The functionality is the same as default_optimizer._eval_acc,
        but used for torchtext.utils.Dataset rather than a torch.data.utils.Dataset
        """
        soft_to_hard_fn, soft_to_hard_fn_kwargs = _validate_soft_to_hard_args(self.soft_to_hard_fn,
                                                                              self.soft_to_hard_fn_kwargs)

        n_correct = defaultdict(int)
        n_total = defaultdict(int)
        model.eval()

        # INPUT: batch of text data + label (batch is gathered with similar lengths with bucket iterator)
        total_val_loss = 0.

        prediction_list_correct = defaultdict(int)
        prediction_list_total = defaultdict(int)
        pred_labels_lst = []
        orig_labels_lst = []
        orig_words_lst = []
        sep_token = self.tokenizer.sep_token
        pad_token = self.tokenizer.pad_token
        cls_token = self.tokenizer.cls_token

        with torch.no_grad():
            for batch_idx, (input_ids, attention_mask, labels, label_mask, valid_ids, token_type_ids) in enumerate(data_loader):

                input_ids = input_ids.to(self.device)
                attention_mask = attention_mask.to(self.device)
                labels = labels.to(self.device)
                label_mask = label_mask.to(self.device)
                valid_ids = valid_ids.to(self.device)
                token_type_ids = token_type_ids.to(self.device)

                # logits = model.predict(input_ids, input_mask=attention_mask)

                # preds: batch size , sequence len, nfeatures
                # labels: batch size, sequence len

                # Compute loss and predictions
                batch_train_loss, logits = model(input_ids,
                                                 token_type_ids=token_type_ids,
                                                 attention_mask=attention_mask,
                                                 labels=labels,
                                                 valid_ids=valid_ids)

                predictions = torch.argmax(logits, dim=2)

                for batch in range(label_mask.shape[0]):
                    orig_words = self.tokenizer.decode(input_ids.data[batch]).split(' ')
                    mask = label_mask.data[batch]
                    preds = predictions.data[batch]
                    lab = labels.data[batch]

                    temp_labels = []
                    temp_preds = []
                    temp_words = []

                    for i, m in enumerate(mask):
                        if m == 1:
                            temp_labels.append(self.id2label[lab[i].item()])
                            temp_preds.append(self.id2label[preds[i].item()])

                    for w in orig_words:
                        if w in [sep_token, pad_token, cls_token]:
                            continue
                        temp_words.append(w)

                    pred_labels_lst.append(temp_preds)
                    orig_labels_lst.append(temp_labels)
                    orig_words_lst.append(temp_words)

                batch_loss = batch_train_loss.item()
                total_val_loss += batch_loss

                running_acc, n_total, n_correct = _running_eval_acc(logits, labels, label_mask,
                                                                    n_total=n_total,
                                                                    n_correct=n_correct,
                                                                    soft_to_hard_fn=soft_to_hard_fn,
                                                                    soft_to_hard_fn_kwargs=soft_to_hard_fn_kwargs)

                if (loss_fn is not None and np.isnan(batch_loss)) or np.isnan(running_acc):
                    # TODO: update to get the actual input text from tokenizer
                    _save_nandata(input_ids, predictions, labels, batch_train_loss, batch_loss,
                                  running_acc, n_total, n_correct, model)

        eval_list = []

        for orig_w, oril, prel in zip(orig_words_lst, orig_labels_lst, pred_labels_lst):

            for ot, ol, pl in zip(orig_w, oril, prel):
                if ot in [self.pad_token, self.cls_token, self.sep_token]:
                    continue
                eval_list.append(f"{ot} {ol} {pl}\n")
            eval_list.append("\n")

        counts = conlleval.evaluate(eval_list)
        report = conlleval.report_notprint(counts)
        for line in report:
            logger.info(line.strip())

        for k in prediction_list_total.keys():
            logger.info('Acc: {} = {}%'.format(k, 0 if prediction_list_total[k] == 0 else (float(
                prediction_list_correct[k]) / float(prediction_list_total[k])) * 100))

        overall, by_type = conlleval.metrics(counts)

        total_val_loss /= float(len(data_loader))

        return running_acc, n_total, n_correct, total_val_loss, counts

    def convert_dataset_to_dataiterator(self, dataset: CSVTextDataset, batch_size: int = None) -> TextDataIterator:
        # NOTE: we use the argument drop_last for the DataLoader (used for the CSVDataset), but no such argument
        # exists for the BucketIterator.  TODO: test whether this might become a problem.

        if not batch_size:
            batch_size_in = self.batch_size
        else:
            batch_size_in = batch_size

        train_sampler = torch.utils.data.RandomSampler(dataset)
        return torch.utils.data.DataLoader(dataset, sampler=train_sampler, batch_size=batch_size_in)

    def train(self, net: torch.nn.Module, dataset: CSVTextDataset,
              torch_dataloader_kwargs: dict = None, use_amp: bool = False) -> (torch.nn.Module, Sequence[EpochStatistics], int):
        """
        Train the network.
        :param net: the network to train
        :param dataset: the dataset to train the network on
        :param torch_dataloader_kwargs: any additional kwargs to pass to PyTorch's native DataLoader
        :param use_amp: if True, uses automated mixed precision for FP16 training.
        :return: the trained network, and a list of EpochStatistics objects which contain the statistics for training,
                and the # of epochs on which the net was trained
        """

        self.ner_per_epoch_report_filepath = os.path.join(self.ner_report_dirpath, 'ner_detailed_stats.' + net.__class__.__name__ + '_' + str(self.run_id) + '.json')
        self.ner_report_filepath = os.path.join(self.ner_report_dirpath, 'ner_stats.' + net.__class__.__name__ + '_' + str(self.run_id) + '.json')

        net = net.to(self.device)

        net.train()  # put network into training mode
        # TODO: Attempt to use AdamW
        if self.optimizer_str == 'adam':
            self.optimizer = optim.Adam(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str == 'sgd':
            self.optimizer = optim.SGD(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str == 'adamw':
            self.optimizer = optim.AdamW(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str not in VALID_OPTIMIZERS:
            msg = self.optimizer_str + " is not a supported optimizer!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            msg = self.optimizer_str + " not yet implemented!"
            logger.error(msg)
            raise NotImplementedError(msg)

        # set according to the following guidelines:
        # https://discuss.pytorch.org/t/when-to-set-pin-memory-to-true/19723
        # https://pytorch.org/docs/stable/notes/cuda.html
        pin_memory = False
        if self.device.type != 'cpu':
            pin_memory = True

        # split into train & validation datasets, and setup data loaders
        data_loader_kwargs_in = dict(batch_size=self.batch_size, pin_memory=pin_memory, drop_last=True, shuffle=True)
        if torch_dataloader_kwargs:
            data_loader_kwargs_in.update(torch_dataloader_kwargs)

        val_dataloader_kwargs_in = dict(batch_size=self.batch_size, pin_memory=pin_memory, drop_last=False, shuffle=False)

        if self.optimizer_cfg.training_cfg.val_dataloader_kwargs is not None:
            val_dataloader_kwargs_in.update(self.optimizer_cfg.training_cfg.val_dataloader_kwargs)
        if torch_dataloader_kwargs:
            val_dataloader_kwargs_in.update(torch_dataloader_kwargs)

        logger.info('DataLoader[Train/Val] kwargs=' + str(torch_dataloader_kwargs))

        train_dataset, val_dataset = train_val_dataset_split(dataset, self.optimizer_cfg.training_cfg.train_val_split,
                                                             self.optimizer_cfg.training_cfg.val_data_transform,
                                                             self.optimizer_cfg.training_cfg.val_label_transform)
        # try to split the val_dataset into clean & triggered
        val_dataset_clean, val_dataset_triggered = split_val_clean_trig(val_dataset)

        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        train_loader = self.convert_dataset_to_dataiterator(train_dataset)
        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        val_clean_loader = self.convert_dataset_to_dataiterator(val_dataset_clean) if \
            len(val_dataset_clean) > 0 else []
        val_triggered_loader = self.convert_dataset_to_dataiterator(val_dataset_triggered) if \
            len(val_dataset_triggered) > 0 else []

        logger.info('#Train[%d]/#ValClean[%d]/#ValTriggered[%d]' %
                    (len(train_loader), len(val_clean_loader), len(val_triggered_loader)))

        num_update_steps_per_epoch = len(train_loader) // self.gradient_accumulation_steps
        num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)
        max_steps = math.ceil(self.num_epochs * num_update_steps_per_epoch)

        self.scheduler = get_linear_schedule_with_warmup(self.optimizer, num_warmup_steps=self.warmup_steps, num_training_steps=max_steps)

        # if self.optimizer_cfg.training_cfg.lr_scheduler is not None:
        #     self.lr_scheduler = self.optimizer_cfg.training_cfg.lr_scheduler(self.optimizer, **self.optimizer_cfg.training_cfg.lr_scheduler_init_kwargs)

        # TODO: Look into adding cyclic scheduler, 1e-4 2.5+/-
        # step_size_up = num_batches / 2
        # base_lr = 1e-5   min = base_lr/5 max = base_lr*5

        # stores training & val data statistics for every epoch
        epoch_stats = []
        best_net = None
        self.best_epoch = 0
        val_loss_array = np.zeros(0, dtype=np.float32)

        num_epochs_to_monitor = 1
        if self.optimizer_cfg.training_cfg.early_stopping:
            num_epochs_to_monitor = self.optimizer_cfg.training_cfg.early_stopping.num_epochs

        epoch = 0
        done = False

        best_acc = 0.0

        while not done:
            train_stats, validation_stats = self.train_epoch(net, train_loader, val_clean_loader, val_triggered_loader,
                                                             epoch, use_amp=use_amp)
            epoch_training_stats = EpochStatistics(epoch, train_stats, validation_stats)
            epoch_stats.append(epoch_training_stats)
            val_loss_array = np.append(val_loss_array, validation_stats.get_val_loss())

            if self.save_best_model:
                # if self.optimizer_cfg.training_cfg.early_stopping:
                #     val_loss_eps = np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps)
                # else:
                #     val_loss_eps = 0
                cur_acc = validation_stats.get_val_acc()
                # loss_threshold = np.min(val_loss_array) + np.abs(val_loss_eps)
                # use validation accuracy as the metric for deciding the best model
                # if validation_stats.get_val_loss() < loss_threshold:
                if best_acc < cur_acc:
                    best_acc = cur_acc
                    self.best_epoch = epoch

                    # msg = "Updating best model with epoch:[%d] loss:[%0.02f] as its less than the best loss plus eps[%0.2e]." % (epoch, validation_stats.get_val_loss(), np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps))
                    msg = "Updating best model with epoch:[%d] acc:[%0.02f] as its better than the best acc[%0.2e]." % (
                        epoch, cur_acc,
                        best_acc)
                    logger.info(msg)
                    best_net = copy.deepcopy(net)

            # early stopping
            # record the val loss of the last batch in the epoch.  if N epochs after the best val_loss, we have not
            # improved the val-loss by at least eps, we quit
            if self.optimizer_cfg.training_cfg.early_stopping:
                # EarlyStoppingConfig validates that eps > 0 as well ..
                error_from_best = np.abs(val_loss_array - np.min(val_loss_array))
                error_from_best[error_from_best < np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps)] = 0
                best_val_loss_epoch = np.where(error_from_best == 0)[0][0]  # unpack numpy array, select first time since that value has happened

                if epoch >= (best_val_loss_epoch + num_epochs_to_monitor):
                    epoch += 1  # we do this b/c of the break to keep the accounting of epoch # returned to
                    # the user to be one based
                    msg = "Exiting training loop in epoch: %d - due to early stopping criterion being met!" % (epoch,)
                    logger.warning(msg)
                    done = True

            epoch += 1
            if self.optimizer_cfg.training_cfg.early_stopping:
                # in case something goes wrong, we exit after training a long time ...
                if epoch >= MAX_EPOCHS:
                    done = True
            else:
                if epoch >= self.num_epochs:
                    done = True

        if self.save_best_model:
            return best_net, epoch_stats, epoch, self.best_epoch
        else:
            return net, epoch_stats, epoch, epoch

    def train_epoch(self, model: nn.Module,
                    train_loader: TextDataIterator,
                    val_clean_loader: TextDataIterator,
                    val_triggered_loader: TextDataIterator,
                    epoch_num: int, progress_bar_disable: bool = False, use_amp: bool = False):
        """
        Runs one epoch of training on the specified model

        :param model: the model to train for one epoch
        :param train_loader: a DataLoader object pointing to the training dataset
        :param val_loader: a DataLoader object pointing to the validation dataset
        :param epoch_num: the epoch number that is being trained
        :param progress_bar_disable: if True, disables the progress bar
        :return: a list of statistics for batches where statistics were computed
        """

        pid = os.getpid()
        train_dataset_len = len(train_loader.dataset)

        scaler = None
        if use_amp:
            scaler = torch.cuda.amp.GradScaler()

        train_n_correct, train_n_total = None, None
        val_n_correct, val_n_total = None, None
        sum_batchmean_train_loss = 0
        running_train_acc = 0
        num_batches = len(train_loader)

        # put network and embedding into training mode
        model.train()

        loop = tqdm(train_loader, disable=self.optimizer_cfg.reporting_cfg.disable_progress_bar)

        for batch_idx, (input_ids, attention_mask, labels, label_mask, valid_ids, token_type_ids) in enumerate(loop):
            # for batch in range(label_mask.shape[0]):
            #     print(self.tokenizer.decode(input_ids.data[batch]))
            # zero out previous gradient computations
            self.optimizer.zero_grad()

            # batch = tuple(t.to(self.device) for t in batch)
            # input_ids, attention_mask, labels, label_mask, valid_ids, token_type_ids = batch
            input_ids = input_ids.to(self.device)
            attention_mask = attention_mask.to(self.device)
            labels = labels.to(self.device)
            label_mask = label_mask.to(self.device)
            valid_ids = valid_ids.to(self.device)
            token_type_ids = token_type_ids.to(self.device)

            if use_amp:
                with torch.cuda.amp.autocast():
                    # predictions = model((input_ids, attention_mask, label_mask, labels))
                    batch_train_loss, predictions = model(input_ids,
                                                          token_type_ids=token_type_ids,
                                                          attention_mask=attention_mask,
                                                          labels=labels,
                                                          valid_ids=valid_ids)
                    # preds: batch size , sequence len, nfeatures
                    # labels: batch size, sequence len
                    # active_preds = predictions.view(-1, predictions.shape[2])
                    # active_labels = labels.view(-1)

                    # compute metrics
                    # batch_train_loss = self._eval_loss_function(active_preds, active_labels)
            else:
                batch_train_loss, predictions = model(input_ids,
                                                      token_type_ids=token_type_ids,
                                                      attention_mask=attention_mask,
                                                      labels=labels,
                                                      valid_ids=valid_ids)

                # predictions = model((input_ids, attention_mask, label_mask))

                # preds: batch size , sequence len, nfeatures
                # labels: batch size, sequence len
                # active_preds = predictions.view(-1, predictions.shape[2])
                # active_labels = labels.view(-1)

                # compute metrics
                # batch_train_loss = self._eval_loss_function(active_preds, active_labels)

            sum_batchmean_train_loss += batch_train_loss.item()
            running_train_acc, train_n_total, train_n_correct = \
                _running_eval_acc(predictions, labels, label_mask, n_total=train_n_total, n_correct=train_n_correct,
                                  soft_to_hard_fn=self.optimizer_cfg.training_cfg.soft_to_hard_fn,
                                  soft_to_hard_fn_kwargs=self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs)

            # backward pass
            if use_amp:
                # Scales loss.  Calls backward() on scaled loss to create scaled gradients.
                # Backward passes under autocast are not recommended.
                # Backward ops run in the same dtype autocast chose for corresponding forward ops.
                scaler.scale(batch_train_loss).backward()
            else:
                if np.isnan(sum_batchmean_train_loss) or np.isnan(running_train_acc):
                    # TODO: Figure out how to pass the original text ... input ids is tokenized
                    _save_nandata(input_ids, predictions, labels, batch_train_loss, sum_batchmean_train_loss, running_train_acc,
                                  train_n_total, train_n_correct, model)
                batch_train_loss.backward()

            # perform gradient clipping if configured
            if self.optimizer_cfg.training_cfg.clip_grad:
                if use_amp:
                    # Unscales the gradients of optimizer's assigned params in-place
                    scaler.unscale_(self.optimizer)

                if self.optimizer_cfg.training_cfg.clip_type == 'norm':
                    # clip_grad_norm_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_norm_(model.parameters(), self.optimizer_cfg.training_cfg.clip_val,
                                                    **self.optimizer_cfg.training_cfg.clip_kwargs)
                elif self.optimizer_cfg.training_cfg.clip_type == 'val':
                    # clip_grad_val_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_value_(
                        model.parameters(), self.optimizer_cfg.training_cfg.clip_val)
                else:
                    msg = "Unknown clipping type for gradient clipping!"
                    logger.error(msg)
                    raise ValueError(msg)
            if use_amp:
                # scaler.step() first unscales the gradients of the optimizer's assigned params.
                # If these gradients do not contain infs or NaNs, optimizer.step() is then called,
                # otherwise, optimizer.step() is skipped.
                scaler.step(self.optimizer)
                # Updates the scale for next iteration.
                scaler.update()
            else:
                self.optimizer.step()

            self.scheduler.step()

            # report batch statistics to tensorflow
            if self.tb_writer:
                try:
                    batch_num = int(epoch_num * num_batches + batch_idx)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-train_loss',
                                              batch_train_loss.item(), global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-running_train_acc',
                                              running_train_acc, global_step=batch_num)
                except:
                    # TODO: catch specific expcetions
                    pass

            loop.set_description('Epoch {}/{}'.format(epoch_num + 1, self.num_epochs))
            loop.set_postfix(avg_train_loss=batch_train_loss.item())

            if batch_idx % self.num_batches_per_logmsg == 0:
                # TODO: Determine best way to get text (possibly from tokenizer if needed...)
                # acc_per_label = "Accuracy Per item: "
                # for k in train_n_total.keys():
                #      acc_per_label += "{}: {}, ".format(k, 0 if train_n_total[k] == 0 else float(train_n_correct[k]) / float(train_n_total[k]))

                logger.info('{}\tTrain Epoch: {} [{}/{} ({:.0f}%)]\tTrainLoss: {:.6f}\tTrainAcc: {:.6f}'.format(
                    pid, epoch_num, batch_idx * input_ids.shape[0], train_dataset_len,
                    # pid, epoch_num, batch_idx * len(text), train_dataset_len,
                                    100. * batch_idx / num_batches, batch_train_loss.item(), running_train_acc))

        train_stats = EpochTrainStatistics(running_train_acc, sum_batchmean_train_loss / float(num_batches))
        clean_counts = None
        triggered_counts = None
        # if we have validation data, we compute on the validation dataset
        num_val_batches_clean = len(val_clean_loader)
        if num_val_batches_clean > 0:
            logger.info('Running Validation on Clean Data')
            running_val_clean_acc, clean_n_total, clean_n_correct, val_clean_loss, clean_counts = \
                self._eval_acc(val_clean_loader, model)
        else:
            logger.info("No dataset computed for validation on clean dataset!")
            running_val_clean_acc = None
            val_clean_loss = None

        num_val_batches_triggered = len(val_triggered_loader)
        if num_val_batches_triggered > 0:
            logger.info('Running Validation on Triggered Data')
            running_val_triggered_acc, triggered_n_total, triggered_n_correct, val_triggered_loss, triggered_counts = \
                self._eval_acc(val_triggered_loader, model)
        else:
            logger.info(
                "No dataset computed for validation on triggered dataset!")
            running_val_triggered_acc = None
            val_triggered_loss = None

        validation_stats = EpochValidationStatistics(running_val_clean_acc, val_clean_loss,
                                                     running_val_triggered_acc, val_triggered_loss)
        if num_val_batches_clean > 0:
            acc_per_label = "{"
            for k in train_n_total.keys():
                acc_per_label += "{}: {}, ".format(k, 0 if clean_n_total[k] == 0 else float(clean_n_correct[k]) / float(clean_n_total[k]))
            acc_per_label += "}"
            logger.info('{}\tTrain Epoch: {} \tCleanValLoss: {:.6f}\tCleanValAcc: {:.6f}\nCleanPerLabelAcc: {}\nclean_total: {}\nclean_correct: {}'.format(
                pid, epoch_num, val_clean_loss, running_val_clean_acc, acc_per_label, clean_n_total, clean_n_correct))
        if num_val_batches_triggered > 0:
            acc_per_label = "{"
            for k in triggered_n_total.keys():
                acc_per_label += "{}: {}, ".format(k, 0 if triggered_n_total[k] == 0 else float(triggered_n_correct[k]) / float(triggered_n_total[k]))
            acc_per_label += "}"
            logger.info('{}\tTrain Epoch: {} \tTriggeredValLoss: {:.6f}\tTriggeredValAcc: {:.6f}\tTriggeredPerLabelAcc: {}'.format(
                pid, epoch_num, val_triggered_loss, running_val_triggered_acc, acc_per_label))

        if self.tb_writer:
            try:
                batch_num = int((epoch_num + 1) * num_batches)
                if num_val_batches_clean > 0:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-clean-val-loss', val_clean_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-clean-val_acc', running_val_clean_acc, global_step=batch_num)
                if num_val_batches_triggered > 0:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-triggered-val-loss', val_triggered_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-triggered-val_acc', running_val_triggered_acc, global_step=batch_num)
            except:
                pass

        # Add epoch to stats
        self.ner_metrics.add_epoch_stats(epoch_num, clean_counts, triggered_counts)

        return train_stats, validation_stats

    def test(self, net: nn.Module, clean_data: CSVTextDataset, triggered_data: CSVTextDataset,
             clean_test_triggered_labels_data: CSVTextDataset,
             torch_dataloader_kwargs: dict = None) -> dict:
        """
        Test the trained network
        :param net: the trained module to run the test data through
        :param clean_data: the clean Dataset
        :param triggered_data: the triggered Dataset, if None, not computed
        :param clean_test_triggered_labels_data: triggered part of the training dataset but with correct labels; see
            DataManger.load_data for more information.
        :param torch_dataloader_kwargs: any keyword arguments to pass directly to PyTorch's DataLoader
        :return: a dictionary of the statistics on the clean and triggered data (if applicable)
        """
        test_data_statistics = {}
        test_data_statistics['clean_metrics'] = None
        test_data_statistics['triggered_metrics'] = None
        test_data_statistics['clean_test_metrics'] = None
        net.eval()

        triggered_counts = None
        clean_test_counts = None
        test_counts = None
        pin_memory = False
        if self.device.type != 'cpu':
            pin_memory = True

        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        data_loader_kwargs_in = dict(batch_size=1, pin_memory=pin_memory, drop_last=True, shuffle=False)
        if torch_dataloader_kwargs:
            data_loader_kwargs_in.update(torch_dataloader_kwargs)
        logger.info('DataLoader[Test] kwargs=' + str(torch_dataloader_kwargs))
        data_loader = self.convert_dataset_to_dataiterator(clean_data, batch_size=1)

        # Test the classification accuracy on clean data only, for all labels.
        test_acc, test_n_total, test_n_correct, _, test_counts = self._eval_acc(data_loader, net)

        acc_per_label = "{"
        for k in test_n_total.keys():
            acc_per_label += "{}: {}, ".format(k, 0 if test_n_total[k] == 0 else float(test_n_correct[k]) / float(
                test_n_total[k]))
        acc_per_label += "}"

        test_data_statistics['clean_accuracy'] = test_acc
        test_data_statistics['clean_n_total'] = test_n_total
        test_data_statistics['clean_per_label_accuracy'] = acc_per_label
        logger.info("Accuracy on clean test data: %0.02f" % (test_data_statistics['clean_accuracy'],))
        logger.info('Per label test accuracy: {}'.format(acc_per_label))
        logger.info('Total per label correct: {}'.format(test_n_correct))
        logger.info('Total per label: {}'.format(test_n_total))

        if triggered_data is not None:
            # Test the classification accuracy on triggered data only, for all labels.
            # we set batch_size=1 b/c
            data_loader = self.convert_dataset_to_dataiterator(triggered_data, batch_size=1)
            test_acc, test_n_total, test_n_correct, _, triggered_counts = self._eval_acc(data_loader, net)
            acc_per_label = "{"
            for k in test_n_total.keys():
                acc_per_label += "{}: {}, ".format(k, 0 if test_n_total[k] == 0 else float(test_n_correct[k]) / float(
                    test_n_total[k]))
            acc_per_label += "}"

            test_data_statistics['triggered_accuracy'] = test_acc
            test_data_statistics['triggered_n_total'] = test_n_total
            test_data_statistics['triggered_per_label_accuracy'] = acc_per_label

            logger.info("Accuracy on triggered test data: %0.02f for n=%s" %
                        (test_data_statistics['triggered_accuracy'], str(test_n_total)))

        if clean_test_triggered_labels_data is not None:
            # Test the classification accuracy on clean data for labels which have corresponding triggered examples.
            # For example, if an MNIST dataset was created with triggered examples only for labels 4 and 5,
            # then this dataset is the subset of data with labels 4 and 5 that don't have the triggers.
            data_loader = self.convert_dataset_to_dataiterator(clean_test_triggered_labels_data, batch_size=1)
            test_acc, test_n_total, test_n_correct, _, clean_test_counts = self._eval_acc(data_loader, net)
            acc_per_label = "{"
            for k in test_n_total.keys():
                acc_per_label += "{}: {}, ".format(k, 0 if test_n_total[k] == 0 else float(test_n_correct[k]) / float(
                    test_n_total[k]))
            acc_per_label += "}"

            test_data_statistics['clean_test_triggered_label_accuracy'] = test_acc
            test_data_statistics['clean_test_triggered_label_n_total'] = test_n_total
            test_data_statistics['clean_test_per_label_accuracy'] = acc_per_label

            logger.info("Accuracy on clean-data-triggered-labels: %0.02f for n=%s" %
                        (test_data_statistics['clean_test_triggered_label_accuracy'], str(test_n_total)))
            logger.info('Per label clean test accuracy: {}'.format(acc_per_label))

        self.ner_metrics.add_best(self.best_epoch, test_counts, clean_test_counts, triggered_counts, )

        # Write the report for ner_metrics
        self.ner_metrics.write_per_epoch(self.ner_per_epoch_report_filepath)
        self.ner_metrics.write_best_epoch(self.ner_report_filepath)

        return test_data_statistics



from trojai.datagen.image_entity import GenericImageEntity, ImageEntity
from trojai.datagen.transform_interface import ImageTransform
from trojai.datagen.image_conversion_utils import normalization_to_rgb, normalization_from_rgb

import wand.image
import wand.color
import wand.drawing
import numpy as np
from numpy.random import RandomState
import cv2

import math
from abc import abstractmethod
from io import BytesIO

import logging
logger = logging.getLogger(__name__)


class FilterXForm(ImageTransform):
    """
    Create filter xform, if no channel order is specified it is assumed to be in BGR order (opencv default),
    this refers only to the first 3 channels of input data as the alpha channel is handled independently
    """
    def __init__(self, channel_order: str = 'BGR', pre_normalize: bool = True, post_normalize: bool = True):
        self.valid_channel_orders = {'BGR', 'RGB'}
        if channel_order not in self.valid_channel_orders:
            raise ValueError("Unknown channel order specified for %s!"
                             "  Valid options are %s!" % (self.__repr__(), str(self.valid_channel_orders)))
        self.channel_order = channel_order
        self.pre_normalize = pre_normalize
        self.post_normalize = post_normalize

    """
    Abstract class containing wand interface functionality common to all filter transforms
    """
    @abstractmethod
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        """
        subclass defined function to be called by do
        :param image: wand Image to be filtered
        :return: filtered wand Image
        """
        pass

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Compresses 3-channel image input image as a specified filetype and stores in memory,
        passes to into wand and applies filter, stores filtered image as specified filetype again in memory,
        which is then decompressed back into 3-channel image
        :param input_obj: entity to be transformed
        :param random_state_obj: object to hold random state and enable reproducibility
        :return:new entity with transform applied
        """
        data = input_obj.get_data()
        original_n_chan = data.shape[2]
        rgb_data, alpha_ch = normalization_to_rgb(data, self.pre_normalize, self.__repr__())
        logger.debug("%s is treating input as %s!" % (self.__repr__(), self.channel_order))
        if self.channel_order == 'RGB':
            rgb_data = cv2.cvtColor(rgb_data, cv2.COLOR_RGB2BGR)

        form = '.bmp'
        success, buffer = cv2.imencode(form, rgb_data)
        # faster than numpy tobytes method
        input_stream = BytesIO(buffer)
        with wand.image.Image(blob=input_stream.getvalue()) as wand_image:
            filtered_wand_image = self.filter(wand_image)
            output_stream = BytesIO()
            filtered_wand_image.save(output_stream)
            rgb_filtered_data = cv2.imdecode(np.frombuffer(output_stream.getbuffer(), np.uint8), 1)

            if self.channel_order == 'RGB':
                rgb_filtered_data = cv2.cvtColor(rgb_filtered_data, cv2.COLOR_BGR2RGB)
            filtered_data = normalization_from_rgb(rgb_filtered_data, alpha_ch, self.post_normalize,
                                                   original_n_chan, self.__repr__())
            return GenericImageEntity(filtered_data, input_obj.get_mask())

    def _colortone(self, image: wand.image.Image, color: str, dst_percent: int, invert: bool) -> None:
        """
        tones either white or black values in image to the provided color,
        intensity of toning depends on dst_percent
        :param image: provided image
        :param color: color to tone image
        :param dst_percent: percentage of image pixel value to include when blending with provided color,
        0 is unchanged, 100 is completely colored in
        :param invert: if True blacks are modified, if False whites are modified
        :return:
        """
        mask_src = image.clone()
        mask_src.colorspace = 'gray'
        if invert:
            mask_src.negate()
        mask_src.alpha_channel = 'copy'

        src = image.clone()
        src.colorize(wand.color.Color(color), wand.color.Color('#FFFFFF'))
        src.composite_channel('alpha', mask_src, 'copy_alpha')

        image.composite_channel('default_channels', src, 'blend',
                                arguments=str(dst_percent) + "," + str(100 - dst_percent))

    def _vignette(self, image: wand.image.Image, color_1: str = 'none', color_2: str = 'black',
                  crop_factor: float = 1.5) -> None:
        """
        applies fading from color_1 to color_2 in radial gradient pattern on given image
        :param image: provided image
        :param color_1: center color
        :param color_2: edge color
        :param crop_factor: size of radial gradient pattern, which is then cropped and combined with image,
        larger values include more of color_1, smaller values include more of color_2
        :return: None
        """
        crop_x = math.floor(image.width * crop_factor)
        crop_y = math.floor(image.height * crop_factor)
        src = wand.image.Image()
        src.pseudo(width=crop_x, height=crop_y, pseudo='radial-gradient:' + color_1 + '-' + color_2)
        src.crop(0, 0, width=image.width, height=image.height, gravity='center')
        src.reset_coords()
        image.composite_channel('default_channels', src, 'multiply')
        image.merge_layers('flatten')


class GothamFilterXForm(FilterXForm):
    """
    Class implementing Instagram's Gotham filter
    """
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        """
        modified from https://github.com/acoomans/instagram-filters/blob/master/instagram_filters/filters/gotham.py
        :param image: provided image
        :return: new filtered image
        """
        filtered_image = image.clone()
        filtered_image.modulate(120, 10, 100)
        filtered_image.colorize(wand.color.Color('#222b6d'), wand.color.Color('#333333'))
        filtered_image.gamma(.9)
        filtered_image.sigmoidal_contrast(True, 3, .5 * filtered_image.quantum_range)
        filtered_image.sigmoidal_contrast(True, 3, .5 * filtered_image.quantum_range)
        return filtered_image


class NashvilleFilterXForm(FilterXForm):
    """
    Class implementing Instagram's Nashville filter
    """
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        """
        modified from https://github.com/acoomans/instagram-filters/blob/master/instagram_filters/filters/nashville.py
        :param image:
        :return: new filtered image
        """
        filtered_image = image.clone()
        self._colortone(filtered_image, '#222b6d', 50, True)
        self._colortone(filtered_image, '#f7daae', 50, False)
        filtered_image.sigmoidal_contrast(True, 3, .5 * filtered_image.quantum_range)
        filtered_image.modulate(100, 150, 100)
        filtered_image.auto_gamma()
        return filtered_image


class KelvinFilterXForm(FilterXForm):
    """
    Class implementing Instagram's Kelvin filter
    """
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        """
        modified from https://github.com/acoomans/instagram-filters/blob/master/instagram_filters/filters/kelvin.py
        :param image: provided image
        :return: new filtered image
        """
        filtered_image = image.clone()
        filtered_image.auto_gamma()
        filtered_image.modulate(120, 50, 100)
        with wand.drawing.Drawing() as draw:
            draw.fill_color = '#FF9900'
            draw.fill_opacity = 0.2
            draw.rectangle(left=0, top=0, width=filtered_image.width, height=filtered_image.height)
            draw(filtered_image)
        return filtered_image


class LomoFilterXForm(FilterXForm):
    """
    Class implementing Instagram's Lomo filter
    """
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        """
        modified from https://github.com/acoomans/instagram-filters/blob/master/instagram_filters/filters/lomo.py
        :param image: provided image
        :return: new filtered image
        """
        filtered_image = image.clone()
        filtered_image.level(.5, channel="R")
        filtered_image.level(.5, channel="G")
        self._vignette(filtered_image)
        return filtered_image


class ToasterXForm(FilterXForm):
    """
    Class implementing Instagram's Toaster filter
    """
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        """
        modified from https://github.com/acoomans/instagram-filters/blob/master/instagram_filters/filters/toaster.py
        :param image: provided image
        :return: new filtered image
        """
        filtered_image = image.clone()
        self._colortone(filtered_image, '#330000', 50, True)
        filtered_image.modulate(150, 80, 100)
        filtered_image.gamma(1.2)
        filtered_image.sigmoidal_contrast(True, 3, .5 * filtered_image.quantum_range)
        filtered_image.sigmoidal_contrast(True, 3, .5 * filtered_image.quantum_range)
        self._vignette(filtered_image, 'none', 'LavenderBlush3')
        self._vignette(filtered_image, '#ff9966', 'none')
        return filtered_image


class NoOpFilterXForm(FilterXForm):
    """
    No operation Transform for testing purposes
    """
    def filter(self, image: wand.image.Image) -> wand.image.Image:
        return image



import torch
import pkg_resources


def make_trojai_model_dict(model):
    """
    Create a TrojAI approved dictionary specification of a PyTorch model for saving to a file. E.g. for a trained model
        'model':
                        save_dict = make_trojai_model_dict(model)
                        torch.save(save_dict, filename)
    :param model: (torch.nn.Module) The desired model to be saved.
    :return: (dict) dictionary containing TrojAI approved information about the model, which can also be used for
        later loading the model.
    """
    # TODO: implement trojai.__version__ and remove pkg_resources module
    return {"class_name": model.__class__.__name__,
            "trojai_version": pkg_resources.require("trojai")[0].version,
            "state_dict": model.state_dict()}


def resave_trojai_model_as_dict(file, new_loc=None):
    """
    Load a fully serialized Pytorch model (i.e. whole model was saved instead of a specification) and save it as a
        TrojAI style dictionary specification.
    :param file: (str) Location of the file to re-save
    :param new_loc: (str) Where to save the file if replacing the original is not desired
    """
    model = torch.load(file)
    model_dict = make_trojai_model_dict(model)
    if new_loc:
        torch.save(model_dict, new_loc)
    else:
        torch.save(model_dict, file)


def clamp(X, l, u, cuda=True):
    """
    Clamps a tensor to lower bound l and upper bound u.
    :param X: the tensor to clamp.
    :param l: lower bound for the clamp.
    :param u: upper bound for the clamp.
    :param cuda: whether the tensor should be on the gpu.
    """

    if type(l) is not torch.Tensor:
        if cuda:
            l = torch.cuda.FloatTensor(1).fill_(l)
        else:
            l = torch.FloatTensor(1).fill_(l)
    if type(u) is not torch.Tensor:
        if cuda:
            u = torch.cuda.FloatTensor(1).fill_(u)
        else:
            u = torch.FloatTensor(1).fill_(u)
    return torch.max(torch.min(X, u), l)


def get_uniform_delta(shape, eps, requires_grad=True):
    """
    Generates a troch uniform random matrix of shape within +-eps.
    :param shape: the tensor shape to create.
    :param eps: the epsilon bounds 0+-eps for the uniform random tensor.
    :param requires_grad: whether the tensor requires a gradient.
    """
    delta = torch.zeros(shape).cuda()
    delta.uniform_(-eps, eps)
    delta.requires_grad = requires_grad
    return delta


from abc import ABC, abstractmethod

from numpy.random import RandomState

from .entity import Entity
from .image_entity import ImageEntity
from .text_entity import TextEntity

"""
Defines a generic Transform object.  
"""


class Transform(ABC):
    """
    A Transform is defined as an operation on an Entity.
    """
    @abstractmethod
    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the specified transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: a random state used to maintain reproducibility through transformations
        :return: the transformed Entity
        """
        pass


class ImageTransform(Transform):
    """
    A Transform specific to ImageEntity objects
    """
    @abstractmethod
    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the specified transformation
        :param input_obj: the input ImageEntity to be transformed
        :param random_state_obj: a random state used to maintain reproducibility through transformations
        :return: the transformed ImageEntity
        """
        pass


class TextTransform(Transform):
    """
    A Transform specific to TextEntity objects
    """
    @abstractmethod
    def do(self, input_obj: TextEntity, random_state_obj: RandomState) -> TextEntity:
        """
        Perform the specified transformation
        :param input_obj: the input TextEntity to be transformed
        :param random_state_obj: a random state used to maintain reproducibility through transformations
        :return: the transformed TextEntity
        """
        pass


# Import packages
from abc import abstractmethod
from pyllist import dllist
import re
from nltk import sent_tokenize

from trojai.datagen.entity import Entity


# Abstract class definition
class TextEntity(Entity):
    @abstractmethod
    def get_delimiters(self):
        pass

    @abstractmethod
    def get_text(self):
        pass

    @abstractmethod
    def __deepcopy__(self, memo):
        pass


# Generic class definition
class GenericTextEntity(TextEntity):
    """
    A class to instantiate an TextEntity for text.
    """

    def __init__(self, input_text: str):
        # Regular expressions
        delimiter_pattern = r'[,:;-]'
        delimiter_regex = re.compile(delimiter_pattern)
        # Create data structures
        self.data = dllist()
        self.delimiters = dllist()
        # Break apart by sentence
        sentences = sent_tokenize(input_text)
        for sentence in sentences:
            # Add a new linked list for the sentence and delimiters
            cur_delimiters = dllist()
            cur_sentence = dllist()
            # Split by whitespace
            t_sentence = sentence.split()
            [cur_sentence.append(word) for word in t_sentence]
            [cur_delimiters.append([index, delimiter_regex.search(word).group(0)])
             for index, word in enumerate(t_sentence) if delimiter_regex.search(word)]
            # Add it in
            self.delimiters.append(cur_delimiters)
            self.data.append(cur_sentence)

    def get_data(self):
        return self.data
    
    def get_delimiters(self):
        return self.delimiters

    def get_text(self):
        # We need to reconstruct the system from the DLL
        return " ".join([" ".join([node for node in sentence]) for sentence in self.data])

    def __deepcopy__(self, memo):
        return GenericTextEntity( self.get_text() )



from trojai.datagen.transform_interface import ImageTransform
from trojai.datagen.image_entity import ImageEntity, GenericImageEntity

from numpy.random import RandomState

import logging
logger = logging.getLogger(__name__)

"""
Defines several transformations related to color manipulation
"""


class GrayscaleRGBToRandomColorXForm(ImageTransform):
    """ Transformer to convert an RGB image to a random color image

    Converts each channel in a 3-channel image to a random color.
    Only pixels which exceed the defined threshold are modified.  This is done for each channel individually.
    """
    def __init__(self, thresh: float = 5) -> None:
        """
        Creates the grayscale-rgb to random color transform object
        :param thresh: threshold above which the sampled R/G/B value will be applied to the input Entity.
        """
        self.thresh = thresh

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual conversion
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: a np.random.RandomState object used to sample the colors and maintain reproducibility
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        if len(img.shape) != 3:
            raise ValueError("Image is not RGB channel - convert first!")

        img[img[:, :, 2].squeeze() > self.thresh, 2] = random_state_obj.choice(255)
        img[img[:, :, 1].squeeze() > self.thresh, 1] = random_state_obj.choice(255)
        img[img[:, :, 0].squeeze() > self.thresh, 0] = random_state_obj.choice(255)

        logger.info("Converted 3-channel Grayscale image to a random color")

        return GenericImageEntity(img, input_obj.get_mask())


class GrayscaleRGBToRandomGrayscaleColorXForm(ImageTransform):
    """
    Converts an RGB grayscale image to an RGB color image by only modifying one channel which is selected randomly,
    where the color gradient of the chosen channel is scaled similarly to the grayscale image
    """
    def __init__(self) -> None:
        """
        Initializes the grayscale-rgb to random color transform object
        """
        pass

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual specified transformation on the input Entity
        :param input_obj: the input object to be transformed
        :param random_state_obj: a np.random.RandomState object which performs the sampling of which channel to modify
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        if len(img.shape) != 3:
            raise ValueError("Image is not RGB channel - convert first!")

        channel = random_state_obj.choice(3) # choose which channel to modify
        # zero out the channels that we don't want color for to produce a
        # grayscale "like" color image
        for ii in range(3):
            if ii != channel:
                img[:, :, ii] = 0

        logger.info("Converted 3-channel Grayscale image with full color applied to channel=%d" % (channel,))

        return GenericImageEntity(img, input_obj.get_mask())



from numpy.random import RandomState
from collections import OrderedDict
from typing import Union

from trojai.datagen.transform_interface import TextTransform
from trojai.datagen.text_entity import TextEntity, GenericTextEntity

import re

import logging
logger = logging.getLogger(__name__)


class ReplacementXForm(TextTransform):
    """
    A wrapper for a word replacement transform, which replaces a specified word in every TextEntity object with a
    specified word.
    """
    def __init__(self, replacements: Union[dict, OrderedDict], ensure_whitespace_surround: bool = False):
        """

        :param replacements: a dictionary or OrderedDict specifying the target word and the replacement word,
                multiple entries in the dictionary mean that multiple replacements will happen.  If a dict is
                specified, it is converted to an OrderedDict such that replacements happen in the original order
                specified.
        :param ensure_whitespace_surround: if True, ensures that the target word to be replaced will only be replaced
            if it is surrounded by whitespace or followed by a delimiter.  Effectively, if enabled, this ensures that
            only whole words are replaced, and not words that are part of other words
        """
        self.replacements = replacements
        self.ensure_whitespace_surround = ensure_whitespace_surround

        self.validate()

    def validate(self):
        if isinstance(self.replacements, dict):
            self.replacements = OrderedDict(self.replacements)
        if isinstance(self.replacements, OrderedDict):
            pass
        else:
            msg = "replacements must be either a dict or OrderedDict object!"
            logger.error(msg)
            raise ValueError(msg)

    def my_replace(self, match):
        """
        A custom string replacement function

        :param match: the input text to replace
        :return: the updated text
        """
        # from: https://stackoverflow.com/a/17730939/1057098
        return self.replacements[match.group(0)]

    def do(self, input_obj: TextEntity, random_state_obj: RandomState) -> TextEntity:
        """
        Performs the transformation

        :param input_obj: the input to be transformed
        :param random_state_obj: random state object used to maintain reproducibility

        Returns: the identity transform of the input (i.e. the input itself)
        """
        text_input = input_obj.get_text()
        if not self.ensure_whitespace_surround:
            updated_text = re.sub('|'.join(r'%s' % re.escape(s) for s in self.replacements),
                                  self.my_replace, text_input)
        else:
            # \b indicates word boundary
            updated_text = re.sub('|'.join(r'\b%s\b' % re.escape(s) for s in self.replacements),
                                  self.my_replace, text_input)

        return GenericTextEntity(updated_text)



# NIST-developed software is provided by NIST as a public service. You may use, copy and distribute copies of the software in any medium, provided that you keep intact this entire notice. You may improve, modify and create derivative works of the software or any portion of the software, and you may copy and distribute such modifications or works. Modified works should carry a notice stating that you changed the software and should note the date and nature of any such change. Please explicitly acknowledge the National Institute of Standards and Technology as the source of the software.

# NIST-developed software is expressly provided "AS IS." NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED, IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.

# You are solely responsible for determining the appropriateness of using and distributing the software and you assume all risks associated with its use, including but not limited to the risks and costs of program errors, compliance with applicable laws, damage to or loss of data, programs or equipment, and the unavailability or interruption of operation. This software is not intended to be used in any situation where a failure could cause risk of injury or damage to property. The software developed by NIST employees is not subject to copyright protection within the United States.

import numpy as np

import PIL.Image
import PIL.ImageDraw
import skimage.io
from functools import reduce
import operator
import math

from .image_entity import ImageEntity


class PolygonTrigger(ImageEntity):
    """
    Defines a programmatically created polygon trigger.
    """
    def __init__(self, img_size: int, n_sides: int, color=None):
        """
        Initializes a polygon object
        :param img_size: the size of the polygons constraining rectangle.
        :param n_sides: the number of polygon sides.
        :param color: polygon color
        """

        if color is not None and len(color) != 3:
            raise RuntimeError('Polygon color must be a 3 elements.')

        self.data = PIL.Image.new('RGBA', (img_size, img_size))
        draw = PIL.ImageDraw.Draw(self.data)

        coords = list()
        for pt in range(n_sides):
            x = np.random.randint(int(0.1 * img_size), int(0.9 * img_size))
            y = np.random.randint(int(0.1 * img_size), int(0.9 * img_size))
            coords.append((x, y))

        center = tuple(map(operator.truediv, reduce(lambda x, y: map(operator.add, x, y), coords), [len(coords)] * 2))
        coords = sorted(coords, key=lambda coord: (-135 - math.degrees(math.atan2(*tuple(map(operator.sub, coord, center))[::-1]))) % 360)

        min_x = img_size
        max_x = 0
        min_y = img_size
        max_y = 0
        for c in coords:
            min_x = min(min_x, c[0])
            max_x = max(max_x, c[0])
            min_y = min(min_y, c[1])
            max_y = max(max_y, c[1])

        x_scale = img_size / (max_x - min_x)
        y_scale = img_size / (max_y - min_y)

        new_coords = list()
        for c in coords:
            new_x = int(x_scale * (c[0] - min_x))
            new_y = int(y_scale * (c[1] - min_y))
            new_coords.append((new_x, new_y))

        coords = new_coords
        if color is None:
            color = [1, 1, 1]
        draw.polygon(coords, fill=(color[0], color[1], color[2], 255))
        # convert from PIL image to numpy array
        self.data = np.array(self.data)
        self.mask = (self.data[:, :, 3] > 0).astype(bool)

    def get_data(self) -> np.ndarray:
        """
        Get the data associated with the ImageEntity
        :return: return a numpy.ndarray representing the image
        """
        return self.data

    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the ImageEntity
        :return: return a numpy.ndarray representing the mask
        """
        return self.mask

    def save(self, output_filepath: str):
        skimage.io.imsave(output_filepath, self.data)





import logging
import os
from typing import Callable, Union
from abc import abstractmethod
import tempfile

import cv2
import pandas as pd
import torch
import torchtext.data
import spacy
from numpy.random import RandomState
from torch.utils.data import Dataset
from tqdm import tqdm

from .data_descriptions import CSVImageDatasetDesc, CSVTextDatasetDesc
from .data_configuration import DEFAULT_LABEL_FIELD_KWARGS, DEFAULT_TEXT_FIELD_KWARGS

logger = logging.getLogger(__name__)

"""
Define some basic default functions for dataset defaults. These allow Dataset objects to be pickled; vs lambda 
functions.
"""


def identity_transform(x):
    return x


def default_image_file_loader(img_loc):
    return torch.from_numpy(cv2.imread(img_loc, cv2.IMREAD_UNCHANGED)).float()


"""
Defines various types of datasets that are used by the DataManager
"""


class DatasetInterface(Dataset):
    def __init__(self, path_to_data: str, *args, **kwargs):
        self.path_to_data = path_to_data

    @abstractmethod
    def set_data_description(self):
        pass

    @abstractmethod
    def get_data_description(self):
        pass


class CSVDataset(DatasetInterface):
    """
    Defines a dataset that is represented by a CSV file with columns "file", "train_label", and optionally
    "true_label". The file column should contain the path to the file that contains the actual data,
    and "train_label" refers to the label with which the data should be trained.  "true_label" refers to the actual
    label of the data point, and can differ from train_label if the dataset is poisoned.  A CSVDataset can support
    any underlying data that can be loaded on the fly and fed into the model (for example: image data)
    """
    def __init__(self, path_to_data: str, csv_filename: str, true_label=False, path_to_csv=None, shuffle=False,
                 random_state: Union[int, RandomState] = None,
                 data_loader: Union[str, Callable] = 'default_image_loader',
                 data_transform=identity_transform, label_transform=identity_transform):
        """
        Initializes a CSVDataset object.
        :param path_to_data: the root folder where the data lives
        :param csv_filename: the CSV file specifying the actual data points
        :param true_label (bool): if True, then use the column "true_label" as the label associated with each
            datapoint.  If False (default), use the column "train_label" as the label associated with each datapoint
        :param path_to_csv: If not None, specifies the folder where the CSV file lives.  If None, it is assumed that
            the CSV file lives in the same directory as the path_to_data
        :param shuffle: if True, the dataset is shuffled before loading into the model
        :param random_state: if specified, seeds the random sampler when shuffling the data
        :param data_loader: either a string value (currently only supports `default_image_loader`), or a callable
            function which takes a string input of the file path and returns the data
        :param data_transform: a callable function which is applied to every data point before it is fed into the
            model. By default, this is an identity operation
        :param label_transform: a callable function which is applied to every label before it is fed into the model.
            By default, this is an identity operation.
        """
        super().__init__(path_to_data)
        if path_to_csv is None:
            path_to_csv = path_to_data
        else:
            path_to_csv = path_to_csv
        self.label = 'train_label'
        if true_label:
            self.label = 'true_label'
        self.data_df = pd.read_csv(os.path.join(path_to_csv, csv_filename))
        self.shuffle = shuffle
        if shuffle:
            self.data_df = self.data_df.sample(frac=1, random_state=random_state).reset_index(drop=True)
        if not callable(data_loader):
            if data_loader == 'default_image_loader':
                self.data_loader = default_image_file_loader
            else:
                msg = "Unknown data loader specified!"
                logger.error(msg)
                raise ValueError(msg)
        else:
            self.data_loader = data_loader
        self.data_transform = data_transform
        self.label_transform = label_transform

        self.data_description = None
        self.set_data_description()

    def __getitem__(self, item):
        data_loc = os.path.join(self.path_to_data, self.data_df.iloc[item]["file"])
        data = self.data_loader(data_loc)
        data = self.data_transform(data)
        label = self.data_df.iloc[item][self.label]
        label = self.label_transform(label)
        return data, label

    def __len__(self):
        return len(self.data_df)

    def get_data_description(self):
        return self.data_description

    def set_data_description(self):
        # set the data description
        num_classes = len(self.data_df[self.label].unique())
        self.data_description = CSVImageDatasetDesc(len(self.data_df), self.shuffle, num_classes)


class CSVTextDataset(torchtext.data.Dataset, DatasetInterface):
    """
    Defines a text dataset that is represented by a CSV file with columns "file", "train_label", and optionally
    "true_label". The file column should contain the path to the file that contains the actual data,
    and "train_label" refers to the label with which the data should be trained.  "true_label" refers to the actual
    label of the data point, and can differ from train_label if the dataset is poisoned.  A CSVTextDataset can support
    text data, and differs from the CSVDataset because it loads all the text data into memory and builds a vocabulary
    from it.
    """
    def __init__(self, path_to_data: str, csv_filename: str, true_label: bool = False,
                 text_field: torchtext.data.Field = None, text_field_kwargs: dict = None,
                 label_field: torchtext.data.LabelField = None, label_field_kwargs: dict = None,
                 shuffle: bool = False, random_state=None, **kwargs):
        """
        Initializes the CSVTextDataset object
        :param path_to_data: root folder where all the data is located
        :param csv_filename: filename of the csv file containing the required fields to load the actual data
        :param true_label (bool): if True, then use the column "true_label" as the label associated with each
        :param text_field (torchtext.data.Field): defines how the text data will be converted to
            a Tensor.  If none, a default will be provided and tokenized with spacy
        :param label_field (torchtext.data.LabelField): defines how to process the label associated with the text
        :param max_vocab_size (int): the maximum vocabulary size that will be built
        :param shuffle: if True, the dataset is shuffled before loading into the model
        :param random_state: if specified, seeds the random sampler when shuffling the data
        :param kwargs: any additional keyword arguments, currently unused

        TODO:
         [ ] - parallelize reading in data from disk
         [ ] - revisit reading entire corpus into memory
         [ ] - loose validation of text_field_kwargs and label_field_kwargs
        """

        if not text_field_kwargs:
            text_field_kwargs = DEFAULT_TEXT_FIELD_KWARGS
        if not label_field_kwargs:
            label_field_kwargs = DEFAULT_LABEL_FIELD_KWARGS

        # try to download the spacy language pack, if the tokenizer is spacy
        if text_field_kwargs['tokenize'] == 'spacy':
            try:
                nlp = spacy.load('en')
            except OSError:
                msg = 'Downloading language model for the spaCy POS tagger'
                logger.warning(msg)
                from spacy.cli import download
                download('en')

        label_column = 'train_label'
        if true_label:
            label_column = 'true_label'
        if text_field is None:
            self.text_field = torchtext.data.Field(**text_field_kwargs)
            msg = "Initialized text_field to default settings!"
            logger.warning(msg)
        else:
            if not isinstance(text_field, torchtext.data.Field):
                msg = "text_field must be of datatype torchtext.data.Field"
                logger.error(msg)
                raise ValueError(msg)
            self.text_field = text_field
        if label_field is None:
            self.label_field = torchtext.data.LabelField(**label_field_kwargs)
            msg = "Initialized label_field to default settings!"
            logger.warning(msg)
        else:
            if not isinstance(label_field, torchtext.data.LabelField):
                msg = "label_field must be of datatype torchtext.data.LabelField"
                logger.error(msg)
                raise ValueError(msg)
            self.label_field = label_field

        fields = [('text', self.text_field), ('label', self.label_field)]
        examples = []

        # read in the specified data into the examples list
        path_to_csv = os.path.join(path_to_data, csv_filename)
        self.data_df = pd.read_csv(path_to_csv)
        if shuffle:
            self.data_df = self.data_df.sample(frac=1, random_state=random_state).reset_index(drop=True)

        for index, row in tqdm(self.data_df.iterrows(), total=len(self.data_df), desc="Loading Text Data..."):
            fname = row['file']
            label = row[label_column]
            with open(os.path.join(path_to_data, fname), 'r') as f:
                z = f.readlines()
                text = ' '.join(z)
            examples.append(torchtext.data.Example.fromlist([text, label], fields))

        self.data_description = None
        super(CSVTextDataset, self).__init__(examples, fields, **kwargs)

    @staticmethod
    def sort_key(ex):
        return len(ex.text)

    def get_data_description(self):
        return self.data_description

    def set_data_description(self):
        # set the data description.  set by build_vocab(...)
        pass

    def build_vocab(self, embedding_vectors_cfg, max_vocab_size, use_vocab=True):
        if use_vocab:
            logger.debug("Building Vocabulary from training data using: " + str(embedding_vectors_cfg) +
                        " with a max vocab size=" + str(max_vocab_size) + " !")
            self.text_field.build_vocab(self,
                                        max_size=max_vocab_size,
                                        vectors=embedding_vectors_cfg,
                                        unk_init=torch.Tensor.normal_)
        self.label_field.build_vocab(self)

        # update the data description
        if use_vocab:
            self.data_description = CSVTextDatasetDesc(vocab_size=len(self.text_field.vocab),
                                                       unk_idx=self.text_field.vocab.stoi[self.text_field.unk_token],
                                                       pad_idx=self.text_field.vocab.stoi[self.text_field.pad_token])
        else:
            self.data_description = CSVTextDatasetDesc(vocab_size=-1,
                                                       unk_idx=0,
                                                       pad_idx=0)


def csv_dataset_from_df(path_to_data, data_df, true_label=False, shuffle=False,
                        random_state: Union[int, RandomState] = None,
                        data_loader: Union[str, Callable] = 'default_image_loader',
                        data_transform=identity_transform, label_transform=identity_transform):
    """
    Initializes a CSVDataset object from a DataFrame rather than a filepath.
    :param path_to_data: root folder where all the data is located
    :param data_df: the dataframe in which the data lives
    :param true_label: (bool) if True, then use the column "true_label" as the label associated with each
    datapoint.  If False (default), use the column "train_label" as the label associated with each datapoint
    :param shuffle: if True, the dataset is shuffled before loading into the model
    :param random_state: if specified, seeds the random sampler when shuffling the data
    :param data_loader: either a string value (currently only supports `default_image_loader`), or a callable
        function which takes a string input of the file path and returns the data
    :param data_transform: a callable function which is applied to every data point before it is fed into the
        model. By default, this is an identity operation
    :param label_transform: a callable function which is applied to every label before it is fed into the model.
        By default, this is an identity operation.
    """
    tmp_file = tempfile.NamedTemporaryFile(suffix='.csv', delete=False)
    tmp_filename = tmp_file.name
    # write df to a temp file
    data_df.to_csv(tmp_filename, sep=',', index=None)
    # load it as a CSVDataset
    csv_dataset = CSVDataset(path_to_data, os.path.basename(tmp_filename), path_to_csv=os.path.dirname(tmp_filename),
                             true_label=true_label, shuffle=shuffle, random_state=random_state, data_loader=data_loader,
                             data_transform=data_transform, label_transform=label_transform)
    # delete tempfile
    os.remove(tmp_filename)
    return csv_dataset


def csv_textdataset_from_df(data_df, true_label: bool = False, text_field: torchtext.data.Field = None,
                            label_field: torchtext.data.LabelField = None,
                            shuffle: bool = False, random_state=None, **kwargs):
    """
    Initializes a CSVDataset object from a DataFrame rather than a filepath.
    :param data_df: the dataframe in which the data lives
    :param true_label: if True, then use the column "true_label" as the label associated with each
    :param text_field: defines how the text data will be converted to
        a Tensor.  If none, a default will be provided and tokenized with spacy
    :param label_field: defines how to process the label associated with the text
    :param max_vocab_size: the maximum vocabulary size that will be built
    :param shuffle: if True, the dataset is shuffled before loading into the model
    :param random_state: if specified, seeds the random sampler when shuffling the data
    :param kwargs: any additional keyword arguments, currently unused
    """
    tmp_file = tempfile.NamedTemporaryFile(suffix='.csv', delete=False)
    tmp_filename = tmp_file.name
    # write df to a temp file
    data_df.to_csv(tmp_filename, sep=',', index=None)
    # load it as a CSVTextDataset
    csv_text_dataset = CSVTextDataset(os.path.dirname(tmp_filename), os.path.basename(tmp_filename),
                                      true_label=true_label, text_field=text_field, label_field=label_field,
                                      shuffle=shuffle, random_state=random_state, **kwargs)
    # delete tempfile
    os.remove(tmp_filename)
    return csv_text_dataset



from trojai.datagen.transform_interface import ImageTransform
from trojai.datagen.image_entity import ImageEntity, GenericImageEntity

import numpy as np
from numpy.random import RandomState
from PIL import ImageEnhance, Image

import logging
logger = logging.getLogger(__name__)


class MMPRMSXForm(ImageTransform):
    """Implements brightness using PIL
    """
    def __init__(self, brightness: float = 1, kwargs: dict = None) -> None:
        """
        Create a scaler object
        :param scale_factor: the scaling amount
        :param kwargs: any keyword arguments to pass to skimge.transform.rescale
        """
        self.brightness = brightness
        if kwargs is None:
            self.kwargs = {}
        else:
            self.kwargs = kwargs

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Performs the scaling on an input Entity using skimage.transform.rescale
        :param input_obj: the input object to be scaled
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        mask = input_obj.get_mask()

        logger.info("Applying %0.02f brightning of image" % (self.brightness,))
        enhancer = ImageEnhance.Sharpness(Image.fromarray(img))
        img_brightned = np.array(enhancer.enhance(self.brightness))

        return GenericImageEntity(img_brightned, mask)



import math

from trojai.datagen.transform_interface import ImageTransform
from trojai.datagen.image_entity import ImageEntity, GenericImageEntity
from trojai.datagen.image_conversion_utils import normalization_to_rgb, normalization_from_rgb
import albumentations.augmentations.transforms as albu

from numpy.random import RandomState

import logging
logger = logging.getLogger(__name__)


"""
A wrapper for all weather related transforms provided by Albumentations project
"""


class BrightenXForm(ImageTransform):
    """Brightens an image by a specified amount.

    """
    def __init__(self, brightness_coeff: float = -1) -> None:
        """
        Initialize the brightener
        :param brightness_coeff: coefficient between 0 and 1 that controls brightness,
        if brightness_coeff is -1, the value is drawn from uniform distribution on [0.0, 1.0) upon each application
        """
        self.brightness_coeff = brightness_coeff
        brightness_coeff_lower, brightness_coeff_upper = brightness_coeff, brightness_coeff
        if self.brightness_coeff == -1:
            brightness_coeff_lower, brightness_coeff_upper = 0.0, 1.0
        self.brighten_object = albu.RandomBrightnessContrast((brightness_coeff_lower, brightness_coeff_upper),
                                                             (0.0, 0.0), always_apply=True)

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        logger.debug("Applying albumentations.RandomBrightnessContrast with coeff=%0.02f" % (self.brightness_coeff,))
        img_xformed = self.brighten_object(random_state=random_state_obj, image=img)['image']

        return GenericImageEntity(img_xformed, input_obj.get_mask())


class DarkenXForm(ImageTransform):
    """ Darkens an image by a specified amount.

    """
    def __init__(self, darkness_coeff: float = -1) -> None:
        """
        Initialize the darkener
        :param darkness_coeff: coefficient between 0 and 1 that controls darkness,
        if brightness_coeff is -1, the value is drawn from the uniform distribution on [0.0, 1.0) upon each
        application
        """
        self.darkness_coeff = darkness_coeff
        darkness_coeff_lower, darkness_coeff_upper = -darkness_coeff, -darkness_coeff
        if self.darkness_coeff == -1:
            darkness_coeff_lower, darkness_coeff_upper = -1.0, 0.0
        self.darken_object = albu.RandomBrightnessContrast((darkness_coeff_lower, darkness_coeff_upper),
                                                           (0.0, 0.0), always_apply=True)

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        logger.debug("Applying albumentations.RandomBrightnessContrast with coeff=%0.02f" % (self.darkness_coeff,))
        img_xformed = self.darken_object(random_state=random_state_obj, image=img)['image']

        return GenericImageEntity(img_xformed, input_obj.get_mask())


class RandomDarkenOrBrightenXForm(ImageTransform):
    """ Randomly brightens/darkens an image by a specified amount

    """
    def __init__(self) -> None:
        """
        Initializes the random brightener/darkener
        """
        self.darken_or_brighten_object = albu.RandomBrightnessContrast((-1.0, 1.0), (0.0, 0.0), always_apply=True)

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        logger.debug("Applying albumentations.RandomBrightnessContrast")
        img_xformed = self.darken_or_brighten_object(random_state=random_state_obj, image=img)['image']

        return GenericImageEntity(img_xformed, input_obj.get_mask())


class AddShadowXForm(ImageTransform):
    """ Adds a shadow to an image

    """
    def __init__(self, no_of_shadows: int = 1, rectangular_roi: tuple = (-1, -1, -1, -1),
                 shadow_dimension: int = 5, pre_normalize: bool = True, post_normalize: bool = True) -> None:
        """
        Initializes the shadow adder
        :param no_of_shadows: the # of shadows
        :param rectangular_roi: region of interest (in pixels) where shadow will be added,
        if rectangular_roi is -1, the roi is chosen uniformly from the lower half of the image upon each application
        :param shadow_dimension: the # of sides to the shadow
        :param pre_normalize: whether to automatically convert input to this transform's required channel format (RGB)
        :param post_normalize: whether to automatically convert output back to its original channel format
        """
        self.no_of_shadows = no_of_shadows
        self.rectangular_roi = rectangular_roi
        self.shadow_dimension = shadow_dimension
        self.shadow_object = None
        self.pre_normalize = pre_normalize
        self.post_normalize = post_normalize

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        original_n_chan = img.shape[2]
        rgb_img, alpha_ch = normalization_to_rgb(img, self.pre_normalize, "AddShadowXForm")

        if self.shadow_object is None:
            no_of_shadows_lower, no_of_shadows_upper = self.no_of_shadows, self.no_of_shadows
            # RandomShadow requires roi to be normalized on [0.0, 1.0]
            roi = ()
            if self.rectangular_roi == (-1, -1, -1, -1):
                roi = (0.0, 0.5, 1.0, 1.0)
            else:
                roi_x1 = self.rectangular_roi[0] / rgb_img.shape[1]
                roi_y1 = self.rectangular_roi[1] / rgb_img.shape[0]
                roi_x2 = self.rectangular_roi[2] / rgb_img.shape[1]
                roi_y2 = self.rectangular_roi[3] / rgb_img.shape[0]
                roi = (roi_x1, roi_y1, roi_x2, roi_y2)
            self.shadow_object = albu.RandomShadow(roi, no_of_shadows_lower, no_of_shadows_upper,
                                                   self.shadow_dimension, always_apply=True)

        logger.debug("Applying albumentations.RandomShadow with shadows=%d, ROI=%s, dimension=%d, pre_normalization=%s,"
                    "post_normalization=%s" %
                    (self.no_of_shadows, str(self.rectangular_roi), self.shadow_dimension, self.pre_normalize,
                     self.post_normalize,))

        rgb_img_xformed = self.shadow_object(random_state=random_state_obj, image=rgb_img)['image']
        img_xformed = normalization_from_rgb(rgb_img_xformed, alpha_ch, self.post_normalize, original_n_chan,
                                             "AddShadowXForm")
        return GenericImageEntity(img_xformed, input_obj.get_mask())


class AddRainXForm(ImageTransform):
    """ Adds rain to an image

    """
    def __init__(self, slant: float = -1, drop_length: float = 20, drop_width: float = 1,
                 drop_color: tuple = (200, 200, 200), rain_type: str = None, pre_normalize: bool = True,
                 post_normalize: bool = True, always_apply: bool = True, probability: float = 0.5):
        """
        Initializes the rain adder
        :param slant:  deviation of rain from normal (-20<=slant<=20),
        if slant is -1, the value will be an integer chosen uniformly on [-10,10) upon each application
        :param drop_length: length of the drop (0<=drop_length<=100)
        :param drop_width: width of the drop (1<=drop_width<=5)
        :param drop_color: color of droplets
        :param rain_type: values in 'drizzle','heavy','torrential', or can be None
        :param pre_normalize: whether to automatically convert input to this transform's required channel format (RGB)
        :param post_normalize: whether to automatically convert output back to its original channel format
        :param always_apply: whether to always apply the transformation, or apply probability percent of the time
        :param probability: probability of the transformation being applied
        """
        self.slant = slant
        slant_lower, slant_upper = slant, slant
        if self.slant == -1:
            slant_lower, slant_upper = -10, 10
        self.drop_length = drop_length
        self.drop_width = drop_width
        self.drop_color = drop_color
        self.rain_type = rain_type
        self.pre_normalize = pre_normalize
        self.post_normalize = post_normalize
        self.rain_object = albu.RandomRain(slant_lower, slant_upper, self.drop_length, self.drop_width,
                                           self.drop_color, blur_value=7, brightness_coefficient=0.7,
                                           rain_type=self.rain_type, always_apply=always_apply, p=probability)

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        original_n_chan = img.shape[2]
        rgb_img, alpha_ch = normalization_to_rgb(img, self.pre_normalize, "AddRainXForm")

        logger.debug("Applying albumentations.RandomRain with slant=%0.02f, drop_length=%0.02f, drop_width=%0.02f,"
                    "drop_color=%s, rain_type=%s, pre_normalization=%s, post_normalization=%s" %
                    (self.slant, self.drop_length, self.drop_width, str(self.drop_color), self.rain_type,
                     self.pre_normalize, self.post_normalize),)

        rgb_img_xformed = self.rain_object(random_state=random_state_obj, image=rgb_img)['image']
        img_xformed = normalization_from_rgb(rgb_img_xformed, alpha_ch, self.post_normalize, original_n_chan,
                                             "AddRainXForm")

        return GenericImageEntity(img_xformed, input_obj.get_mask())


class AddSnowXForm(ImageTransform):
    """ Adds snow to the image

    """
    def __init__(self, snow_coeff: float = -1, pre_normalize: bool = True, post_normalize: bool = True, always_apply: bool =True, probability: float =0.5):
        """
        Initializes the snow adder
        :param snow_coeff: coefficient between 0 and 1 controlling the amount of snow,
        if snow_coeff is -1, the value is drawn from the uniform distribution on [0.0, 1.0) upon each application
        :param pre_normalize: whether to automatically convert input to this transform's required channel format (RGB)
        :param post_normalize: whether to automatically convert output back to its original channel format
        :param always_apply: whether to always apply the transformation, or apply probability percent of the time
        :param probability: probability of the trasformation being applied
        """
        self.snow_coeff = snow_coeff
        snow_coeff_lower, snow_coeff_upper = snow_coeff, snow_coeff
        if self.snow_coeff == -1:
            snow_coeff_lower, snow_coeff_upper = 0.0, 1.0
        self.pre_normalize = pre_normalize
        self.post_normalize = post_normalize

        self.snow_object = albu.RandomSnow(snow_coeff_lower, snow_coeff_upper, brightness_coeff=2.5, always_apply=always_apply, p=probability)

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        original_n_chan = img.shape[2]
        rgb_img, alpha_ch = normalization_to_rgb(img, self.pre_normalize, "AddSnowXForm")

        logger.debug("Applying albumentations.RandomSnow with coeff=%0.02f, pre_normalize=%s, post_normalize=%s" %
                    (self.snow_coeff, self.pre_normalize, self.post_normalize,))

        rgb_img_xformed = self.snow_object(random_state=random_state_obj, image=rgb_img)['image']
        img_xformed = normalization_from_rgb(rgb_img_xformed, alpha_ch, self.post_normalize, original_n_chan,
                                             "AddSnowXForm")
        return GenericImageEntity(img_xformed, input_obj.get_mask())


class AddFogXForm(ImageTransform):
    """ Adds Fog to the image

    """
    def __init__(self, fog_coeff: float = -1, pre_normalize: bool = True, post_normalize: bool = True, always_apply: bool = True, probability: float = 0.5):
        """
        Initializes the fog adder
        :param fog_coeff: coefficient between 0 and 1 controlling the amount of fog,
        if fog_coeff is -1, the value is drawn from the uniform distribution on [0.3, 1.0) upon each application`
        :param pre_normalize: whether to automatically convert input to this transform's required channel format (RGB)
        :param post_normalize: whether to automatically convert output back to its original channel format
        :param always_apply: whether to always apply the transformation, or apply probability percent of the time
        :param probability: probability of the transformation being applied
        """
        self.fog_coeff = fog_coeff
        fog_coeff_lower = fog_coeff
        fog_coeff_upper = fog_coeff
        if fog_coeff == -1:
            fog_coeff_lower, fog_coeff_upper = 0.3, 1.0
        self.pre_normalize = pre_normalize
        self.post_normalize = post_normalize

        self.fog_object = albu.RandomFog(fog_coeff_lower, fog_coeff_upper, alpha_coef=0.08, always_apply=always_apply, p=probability)

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        original_n_chan = img.shape[2]
        rgb_img, alpha_ch = normalization_to_rgb(img, self.pre_normalize, "AddFogXForm")

        logger.debug("Applying albumentations.RandomFog fog with coef=%0.02f, pre_normalize=%s, post_normalize=%s" %
                    (self.fog_coeff, self.pre_normalize, self.post_normalize))

        rgb_img_xformed = self.fog_object(random_state=random_state_obj, image=rgb_img)['image']
        img_xformed = normalization_from_rgb(rgb_img_xformed, alpha_ch, self.post_normalize, original_n_chan,
                                             "AddFogXForm")
        return GenericImageEntity(img_xformed, input_obj.get_mask())


# TODO: explore other ways to handle differences between trojai absolute location and albu relative location for images
class AddSunFlareXForm(ImageTransform):
    """ Adds a Sun Flare to the image

    """
    def __init__(self, flare_center: tuple = (-1, -1), angle: float = -1, no_of_flare_circles: int = 4,
                 src_radius: int = 200,  src_color: tuple = (255, 255, 255), pre_normalize: bool = True,
                 post_normalize: bool = True):
        """
        Initialize the sun flare adder
        :param flare_center: center coordinates (x,y) of the source,
        if flare_center is (-1,-1) the center is chosen uniformly from the upper half of the image upon each application
        :param angle: angle of flare in radians,
        if angle is -1, the normalized value is drawn from the uniform distribution on [0.0, 1.0] upon each application
        :param no_of_flare_circles:  no. of secondary flare circles
        :param src_radius: radius of the primary flare source (in pixels)
        :param src_color: rgb color of the flare source and secondary circles
        :param pre_normalize: whether to automatically convert input to this transform's required channel format (RGB)
        :param post_normalize: whether to automatically convert output back to its original channel format
        """
        self.flare_center = flare_center
        self.angle = angle
        self.no_of_flare_circles = no_of_flare_circles
        self.src_radius = src_radius
        self.src_color = src_color
        self.sunflare_object = None
        self.pre_normalize = pre_normalize
        self.post_normalize = post_normalize

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual defined transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        original_n_chan = img.shape[2]
        rgb_img, alpha_ch = normalization_to_rgb(img, self.pre_normalize, "AddSunFlareXForm")

        # RandomSunFlare requires center and angle of flare to be normalized on [0.0, 1.0],
        if self.sunflare_object is None:
            roi = ()
            if self.flare_center == (-1, -1):
                roi = (0.0, 0.0, 1.0, 0.5)
            else:
                roi_x = self.flare_center[0] / rgb_img.shape[1]
                roi_y = self.flare_center[1] / rgb_img.shape[0]
                roi = (roi_x, roi_y, roi_x, roi_y)
            angle_lower = self.angle / (2 * math.pi)
            angle_upper = self.angle / (2 * math.pi)
            if self.angle == -1:
                angle_lower, angle_upper = 0.0, 1.0
            self.sunflare_object = albu.RandomSunFlare(roi, angle_lower, angle_upper, self.no_of_flare_circles-1,
                                                       self.no_of_flare_circles+1, self.src_radius, self.src_color,
                                                       always_apply=True)
        logger.debug("Applying albumentations.RandomSunFlare with center=%s, angle=%0.02f, # flare-circles=%d,"
                    "flare-radius=%d, color=%s, pre_normalize=%s, post_normalize=%s" %
                    (str(self.flare_center), self.angle, self.no_of_flare_circles,
                     self.src_radius, str(self.src_color), self.pre_normalize, self.post_normalize))

        rgb_img_xformed = self.sunflare_object(random_state=random_state_obj, image=rgb_img)['image']
        img_xformed = normalization_from_rgb(rgb_img_xformed, alpha_ch, self.post_normalize, original_n_chan,
                                             "AddSunFlareXForm")
        return GenericImageEntity(img_xformed, input_obj.get_mask())



# Python version of the evaluation script from CoNLL'00-
# Originates from: https://github.com/spyysalo/conlleval.py

# The MIT License (MIT)
#
# Copyright (c) 2016 Sampo Pyysalo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# NIST-developed software is provided by NIST as a public service. You may use, copy and distribute copies of the software in any medium, provided that you keep intact this entire notice. You may improve, modify and create derivative works of the software or any portion of the software, and you may copy and distribute such modifications or works. Modified works should carry a notice stating that you changed the software and should note the date and nature of any such change. Please explicitly acknowledge the National Institute of Standards and Technology as the source of the software.
# NIST-developed software is expressly provided "AS IS." NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED, IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
# You are solely responsible for determining the appropriateness of using and distributing the software and you assume all risks associated with its use, including but not limited to the risks and costs of program errors, compliance with applicable laws, damage to or loss of data, programs or equipment, and the unavailability or interruption of operation. This software is not intended to be used in any situation where a failure could cause risk of injury or damage to property. The software developed by NIST employees is not subject to copyright protection within the United States.


# Intentional differences:
# - accept any space as delimiter by default
# - optional file argument (default STDIN)
# - option to set boundary (-b argument)
# - LaTeX output (-l argument) not supported
# - raw tags (-r argument) not supported

# add function :evaluate(predicted_label, ori_label): which will not read from file

import sys
import re
import codecs
from collections import defaultdict, namedtuple

ANY_SPACE = '<SPACE>'


class FormatError(Exception):
    pass

Metrics = namedtuple('Metrics', 'tp fp fn prec rec fscore')


class EvalCounts(object):
    def __init__(self):
        self.correct_chunk = 0    # number of correctly identified chunks
        self.correct_tags = 0     # number of correct chunk tags
        self.found_correct = 0    # number of chunks in corpus
        self.found_guessed = 0    # number of identified chunks
        self.token_counter = 0    # token counter (ignores sentence breaks)

        # counts by type
        self.t_correct_chunk = defaultdict(int)
        self.t_found_correct = defaultdict(int)
        self.t_found_guessed = defaultdict(int)


def parse_args(argv):
    import argparse
    parser = argparse.ArgumentParser(
        description='evaluate tagging results using CoNLL criteria',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    arg = parser.add_argument
    arg('-b', '--boundary', metavar='STR', default='-X-',
        help='sentence boundary')
    arg('-d', '--delimiter', metavar='CHAR', default=ANY_SPACE,
        help='character delimiting items in input')
    arg('-o', '--otag', metavar='CHAR', default='O',
        help='alternative outside tag')
    arg('file', nargs='?', default=None)
    return parser.parse_args(argv)


def parse_tag(t):
    m = re.match(r'^([^-]*)-(.*)$', t)
    return m.groups() if m else (t, '')


def evaluate(iterable, options=None):
    if options is None:
        options = parse_args([])    # use defaults

    counts = EvalCounts()
    num_features = None       # number of features per line
    in_correct = False        # currently processed chunks is correct until now
    last_correct = 'O'        # previous chunk tag in corpus
    last_correct_type = ''    # type of previously identified chunk tag
    last_guessed = 'O'        # previously identified chunk tag
    last_guessed_type = ''    # type of previous chunk tag in corpus

    for line in iterable:
        line = line.rstrip('\r\n')

        if options.delimiter == ANY_SPACE:
            features = line.split()
        else:
            features = line.split(options.delimiter)

        if num_features is None:
            num_features = len(features)
        elif num_features != len(features) and len(features) != 0:
            raise FormatError('unexpected number of features: %d (%d)' %
                              (len(features), num_features))

        if len(features) == 0 or features[0] == options.boundary:
            features = [options.boundary, 'O', 'O']
        if len(features) < 3:
            raise FormatError('unexpected number of features in line %s' % line)

        guessed, guessed_type = parse_tag(features.pop())
        correct, correct_type = parse_tag(features.pop())
        first_item = features.pop(0)

        if first_item == options.boundary:
            guessed = 'O'

        end_correct = end_of_chunk(last_correct, correct,
                                   last_correct_type, correct_type)
        end_guessed = end_of_chunk(last_guessed, guessed,
                                   last_guessed_type, guessed_type)
        start_correct = start_of_chunk(last_correct, correct,
                                       last_correct_type, correct_type)
        start_guessed = start_of_chunk(last_guessed, guessed,
                                       last_guessed_type, guessed_type)

        if in_correct:
            if (end_correct and end_guessed and
                last_guessed_type == last_correct_type):
                in_correct = False
                counts.correct_chunk += 1
                counts.t_correct_chunk[last_correct_type] += 1
            elif (end_correct != end_guessed or guessed_type != correct_type):
                in_correct = False

        if start_correct and start_guessed and guessed_type == correct_type:
            in_correct = True

        if start_correct:
            counts.found_correct += 1
            counts.t_found_correct[correct_type] += 1
        if start_guessed:
            counts.found_guessed += 1
            counts.t_found_guessed[guessed_type] += 1
        if first_item != options.boundary:
            if correct == guessed and guessed_type == correct_type:
                counts.correct_tags += 1
            counts.token_counter += 1

        last_guessed = guessed
        last_correct = correct
        last_guessed_type = guessed_type
        last_correct_type = correct_type

    if in_correct:
        counts.correct_chunk += 1
        counts.t_correct_chunk[last_correct_type] += 1

    return counts



def uniq(iterable):
  seen = set()
  return [i for i in iterable if not (i in seen or seen.add(i))]


def calculate_metrics(correct, guessed, total):
    tp, fp, fn = correct, guessed-correct, total-correct
    p = 0 if tp + fp == 0 else 1.*tp / (tp + fp)
    r = 0 if tp + fn == 0 else 1.*tp / (tp + fn)
    f = 0 if p + r == 0 else 2 * p * r / (p + r)
    return Metrics(tp, fp, fn, p, r, f)


def metrics(counts):
    c = counts
    overall = calculate_metrics(
        c.correct_chunk, c.found_guessed, c.found_correct
    )
    by_type = {}
    for t in uniq(list(c.t_found_correct) + list(c.t_found_guessed)):
        by_type[t] = calculate_metrics(
            c.t_correct_chunk[t], c.t_found_guessed[t], c.t_found_correct[t]
        )
    return overall, by_type


def report(counts, out=None):
    if out is None:
        out = sys.stdout

    overall, by_type = metrics(counts)

    c = counts
    out.write('processed %d tokens with %d phrases; ' %
              (c.token_counter, c.found_correct))
    out.write('found: %d phrases; correct: %d.\n' %
              (c.found_guessed, c.correct_chunk))

    if c.token_counter > 0:
        out.write('accuracy: %6.2f%%; ' %
                  (100.*c.correct_tags/c.token_counter))
        out.write('precision: %6.2f%%; ' % (100.*overall.prec))
        out.write('recall: %6.2f%%; ' % (100.*overall.rec))
        out.write('FB1: %6.2f\n' % (100.*overall.fscore))

    for i, m in sorted(by_type.items()):
        out.write('%17s: ' % i)
        out.write('precision: %6.2f%%; ' % (100.*m.prec))
        out.write('recall: %6.2f%%; ' % (100.*m.rec))
        out.write('FB1: %6.2f  %d\n' % (100.*m.fscore, c.t_found_guessed[i]))


def report_notprint(counts, out=None):
    if out is None:
        out = sys.stdout

    overall, by_type = metrics(counts)

    c = counts
    final_report = []
    line = []
    line.append('processed %d tokens with %d phrases; ' %
              (c.token_counter, c.found_correct))
    line.append('found: %d phrases; correct: %d.\n' %
              (c.found_guessed, c.correct_chunk))
    final_report.append("".join(line))

    if c.token_counter > 0:
        line = []
        line.append('accuracy: %6.2f%%; ' %
                  (100.*c.correct_tags/c.token_counter))
        line.append('precision: %6.2f%%; ' % (100.*overall.prec))
        line.append('recall: %6.2f%%; ' % (100.*overall.rec))
        line.append('FB1: %6.2f\n' % (100.*overall.fscore))
        final_report.append("".join(line))

    for i, m in sorted(by_type.items()):
        line = []
        line.append('%17s: ' % i)
        line.append('precision: %6.2f%%; ' % (100.*m.prec))
        line.append('recall: %6.2f%%; ' % (100.*m.rec))
        line.append('FB1: %6.2f  %d\n' % (100.*m.fscore, c.t_found_guessed[i]))
        final_report.append("".join(line))
    return final_report


def end_of_chunk(prev_tag, tag, prev_type, type_):
    # check if a chunk ended between the previous and current word
    # arguments: previous and current chunk tags, previous and current types
    chunk_end = False

    if prev_tag == 'E': chunk_end = True
    if prev_tag == 'S': chunk_end = True

    if prev_tag == 'B' and tag == 'B': chunk_end = True
    if prev_tag == 'B' and tag == 'S': chunk_end = True
    if prev_tag == 'B' and tag == 'O': chunk_end = True
    if prev_tag == 'I' and tag == 'B': chunk_end = True
    if prev_tag == 'I' and tag == 'S': chunk_end = True
    if prev_tag == 'I' and tag == 'O': chunk_end = True

    if prev_tag != 'O' and prev_tag != '.' and prev_type != type_:
        chunk_end = True

    # these chunks are assumed to have length 1
    if prev_tag == ']': chunk_end = True
    if prev_tag == '[': chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    # check if a chunk started between the previous and current word
    # arguments: previous and current chunk tags, previous and current types
    chunk_start = False

    if tag == 'B': chunk_start = True
    if tag == 'S': chunk_start = True

    if prev_tag == 'E' and tag == 'E': chunk_start = True
    if prev_tag == 'E' and tag == 'I': chunk_start = True
    if prev_tag == 'S' and tag == 'E': chunk_start = True
    if prev_tag == 'S' and tag == 'I': chunk_start = True
    if prev_tag == 'O' and tag == 'E': chunk_start = True
    if prev_tag == 'O' and tag == 'I': chunk_start = True

    if tag != 'O' and tag != '.' and prev_type != type_:
        chunk_start = True

    # these chunks are assumed to have length 1
    if tag == '[': chunk_start = True
    if tag == ']': chunk_start = True

    return chunk_start


def return_report(input_file):
    with codecs.open(input_file, "r", "utf8") as f:
        counts = evaluate(f)
    return report_notprint(counts)


def main(argv):
    args = parse_args(argv[1:])

    if args.file is None:
        counts = evaluate(sys.stdin, args)
    else:
        with open(args.file) as f:
            counts = evaluate(f, args)
    report(counts)

if __name__ == '__main__':
    sys.exit(main(sys.argv))


from trojai.datagen.merge_interface import ImageMerge
from trojai.datagen.image_entity import ImageEntity, GenericImageEntity
import trojai.datagen.lighting_utils as lighting_utils

from typing import Callable
import blend_modes
import cv2
import numpy as np
from numpy.random import RandomState
from PIL import Image

import logging
logger = logging.getLogger(__name__)


"""
Module which defines several blend style merge operations.
"""


class GrainMerge(ImageMerge):
    """ Performs a "grain" merge, similar to grain-merge in PhotoShop.

    See here for details:
    https://pythonhosted.org/blend_modes/blend_modes.html?highlight=grain_merge#blend_modes.blend_modes.grain_merge
    """
    def __init__(self, opacity: float = 1) -> None:
        """
        Creates the GrainMerge object
        :param opacity: Desired opacity of layer for blending
        """
        self.opacity = opacity

    def do(self, img_obj: ImageEntity, pattern_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual blend operation on an input and pattern
        :param img_obj: image to be blended upon
        :param pattern_obj: the layer to be blended with the image
        :param random_state_obj: ignored
        :return: the merged object
        """
        img = img_obj.get_data()
        img_mask = img_obj.get_mask()
        pattern = pattern_obj.get_data()
        pattern_mask = pattern_obj.get_mask()
        logger.debug("Grain Merging img w/ shape=%s and pattern w/ shape=%s with opacity=%0.02f",
                    (str(img.shape), str(pattern.shape), self.opacity))
        img_out = blend_modes.grain_merge(img.astype(float), pattern.astype(float), self.opacity)
        mask_out = img_mask & pattern_mask
        return GenericImageEntity(img_out, mask_out)


class AddMerge(ImageMerge):
    """ Merge objects by element wise addition

    """
    def __init__(self) -> None:
        """
        Create the AddMerge object
        """
        pass

    def do(self, img_obj: ImageEntity, pattern_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual Merge operation on the input and pattern
        :param img_obj: image to be added
        :param pattern_obj: pattern to be added
        :param random_state_obj: ignored
        :return: the merged object
        """
        logger.debug("Add Merging img w/ shape=%s and pattern w/ shape=%s",
                    (str(img_obj.get_data().shape), str(pattern_obj.get_data().shape)))
        img_out = cv2.add(img_obj.get_data(), pattern_obj.get_data())
        # TODO: revisit whether this is the correct behavior for the mask
        mask_out = cv2.add(img_obj.get_mask(), pattern_obj.get_mask())
        return GenericImageEntity(img_out, mask_out)


class GrainMergePaste(ImageMerge):
    """ Class which implements a Grain Merge and Paste operation in serial on two Entities.

    See: https://pythonhosted.org/blend_modes/blend_modes.html?highlight=grain_merge#blend_modes.blend_modes.grain_merge
         https://pillow.readthedocs.io/en/3.1.x/reference/Image.html#PIL.Image.Image.paste
    """
    def __init__(self, opacity: float = 1) -> None:
        """
        Creates a GrainMergeAndPaste object with a configurable opacity
        :param opacity: Desired opacity of layer for blending
        """
        self.opacity = opacity

    def do(self, img_obj: ImageEntity, pattern_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual blending operations
        :param img_obj: image to be blended upon
        :param pattern_obj: the layer to be blended with the image & pasted
        :param random_state_obj: ignored
        :return: the merged object
        """
        img = img_obj.get_data()
        img_mask = img_obj.get_mask()
        pattern = pattern_obj.get_data()
        pattern_mask = pattern_obj.get_mask()

        logger.debug("Grain Merging img w/ shape={} and pattern w/ shape={} with opacity={:0.02f}".format(str(img.shape), str(pattern.shape), self.opacity))

        img_r, img_c, _ = img.shape
        pat_r, pat_c, _ = pattern.shape
        if pat_r > img_r or pat_c > img_c:
            msg = "Pattern to be merged into image is larger than the image!"
            logger.error(msg)
            raise ValueError(msg)
        if pat_r < img_r or pat_c < img_c:
            # TODO: make this an option so that we have multiple resize options
            logger.debug("Resizing pattern to match image size with image background!")
            pattern_resized = img.copy()
            row_insert_idx = (img_r-pat_r)//2
            col_insert_idx = (img_c-pat_c)//2
            pattern_resized[row_insert_idx:row_insert_idx+pat_r, col_insert_idx:col_insert_idx+pat_c,:] = pattern
            pattern = pattern_resized.copy()
            pattern_mask_resized = np.zeros((img_r, img_c), dtype=bool)
            pattern_mask_resized[row_insert_idx:row_insert_idx+pat_r, col_insert_idx:col_insert_idx+pat_c] = pattern_mask
            pattern_mask = pattern_mask_resized.copy()

        blended_img = blend_modes.grain_merge(img.astype(float), pattern.astype(float), self.opacity)
        blended_img_raw = Image.fromarray(blended_img.astype(np.uint8))
        pattern_raw = Image.fromarray(pattern.astype(np.uint8))
        logger.debug("Pasting pattern into grain merged image")
        blended_img_raw.paste(pattern_raw, (0, 0), pattern_raw)
        final_img = np.array(blended_img_raw)

        # TODO: revisit whether this is the correct/intended behavior for the mask
        final_mask = img_mask & pattern_mask

        return GenericImageEntity(final_img, final_mask)


class BrightnessAdjustGrainMergePaste(ImageMerge):
    """ Class which implements a brightness adjustment before grain merging & pasting

    """
    def __init__(self, opacity: float = 1,
                 lighting_adjuster: Callable[[np.ndarray, np.ndarray], np.ndarray] =
                 lighting_utils.adjust_brightness_mmavg) -> None:
        """
        Create an instance of the brighness adjustment + grain merging + pasting object
        :param opacity: Desired opacity of layer for blending
        :param lighting_adjuster: A function handle for the lighting adjustment operation
        """
        self.opacity = opacity
        self.lighting_adjuster = lighting_adjuster

    def do(self, img_obj: ImageEntity, pattern_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual blending operations
        :param img_obj: image to be blended upon
        :param pattern_obj: the layer to be blended with the brightness adjusted & image  & pasted
        :param random_state_obj: ignored
        :return: the merged object
        """
        img = img_obj.get_data()
        pattern = pattern_obj.get_data()
        pattern_mask = pattern_obj.get_mask()

        # adjust brightness of pattern to match image
        logger.debug("Adjusting brightness according to:" + str(self.lighting_adjuster))
        pattern_adjusted = self.lighting_adjuster(pattern, img)
        pattern_adjusted_obj = GenericImageEntity(pattern_adjusted, pattern_mask)
        logger.debug("Performing GrainMergePaste with opacity = %0.02f", (self.opacity))
        merger = GrainMergePaste(self.opacity)
        return merger.do(img_obj, pattern_adjusted_obj, random_state_obj)



import cv2
import numpy as np
from numpy.random import RandomState

from trojai.datagen.transform_interface import ImageTransform
from trojai.datagen.image_entity import ImageEntity, GenericImageEntity

import logging
logger = logging.getLogger(__name__)

"""
Defines several transformations related to color manipulation
"""


class PoissonNoiseXForm(ImageTransform):
    """
    Inserts Poisson noise into the image object
    """
    def __init__(self, exponent_base: float = 2.05) -> None:
        """
        Initializes the Poisson Noise Inserter Object
        :param exponent_base:
        """
        self.exponent_base = exponent_base

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual noise insertion
        :param input_obj: the input Entity to which noise is to be added
        :param random_state_obj: a RandomState object used to sample the noise
        :return: the transformed (noise added) Entity
        """
        img = input_obj.get_data()
        vals = len(np.unique(img))
        vals = self.exponent_base ** np.ceil(np.log2(vals))
        noisy = random_state_obj.poisson(img * vals) / float(vals)

        logger.debug("Added Poisson Noise to Image")
        return GenericImageEntity(noisy, input_obj.get_mask())


class GaussianNoiseXForm(ImageTransform):
    """
    Inserts Gaussian noise into the image object
    """
    def __init__(self, mean: float = 0, var: float = 0.5) -> None:
        """
        Initializes the Gaussian Noise Inserter Object
        :param mean: noise mean value
        :param var: noise standard deviation
        :param seed: random seed value
        """
        self.mean = mean
        self.var = var

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual noise insertion
        :param input_obj: the input Entity to which noise is to be added
        :param random_state_obj: a RandomState object used to sample the noise
        :return: the transformed (noise added) Entity
        """
        img = input_obj.get_data()
        row, col, ch = img.shape
        sigma = self.var ** 0.5
        gaussian_noise = random_state_obj.normal(self.mean, sigma, (row, col, ch))
        gaussian_noise = gaussian_noise.reshape(row, col, ch)
        noisy = img + gaussian_noise

        logger.debug("Added Gaussian Noise to Image")
        return GenericImageEntity(noisy, input_obj.get_mask())


class GaussianBlurXForm(ImageTransform):
    """
    Performs Gaussian blurring of an Entity
    """
    def __init__(self, ksize: int = 5, sigmaX: float = 0, sigmaY: float = 0) -> None:
        """
        Initializes the Gaussian Blur Object
        :param ksize: kernel size
        :param sigmaX: x sigma value (leave as 0 to define based on kernel size)
        :param sigmaY: y sigma value (leave as 0 to define based on kernel size)
        """
        if ksize%2 == 0:
            raise ValueError("Kernel size must be odd!")
        self.ksize = ksize
        self.sigmaX = sigmaX
        self.sigmaY = sigmaY

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual blurring operation
        :param input_obj: the input Entity to which noise is to be added
        :param random_state_obj: ignored
        :return: the transformed (noise added) Entity
        """
        img = input_obj.get_data()
        blurred = cv2.GaussianBlur(img, (self.ksize, self.ksize), self.sigmaX, self.sigmaY)

        logger.debug("Added Gaussian Blur w/ Kernel Size=%d to Image" % (self.ksize,))
        return GenericImageEntity(blurred, input_obj.get_mask())


class RandomGaussianBlurXForm(ImageTransform):
    """
    Performs Gaussian blurring of an Entity
    """
    def __init__(self, ksize_min: int = 0, ksize_max: int = 5, sigmaX: float = 0, sigmaY: float = 0) -> None:
        """
        Initializes the Gaussian Blur Object
        :param ksize_min: minimum kernel size to select and apply
        :param ksize_max: maximum kernel size to select and apply
        :param sigmaX: x sigma value (leave as 0 to define based on kernel size)
        :param sigmaY: y sigma value (leave as 0 to define based on kernel size)
        """

        self.ksize_min = int(max(ksize_min - 1, 0) / 2)
        self.ksize_max = int(max(ksize_max - 1, 0) / 2)
        self.sigmaX = sigmaX
        self.sigmaY = sigmaY

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual blurring operation
        :param input_obj: the input Entity to which noise is to be added
        :param random_state_obj: ignored
        :return: the transformed (noise added) Entity
        """

        # select a kernel size
        ksize = int(random_state_obj.randint(int(self.ksize_min), int(self.ksize_max)))
        ksize = int(ksize * 2 + 1)
        if ksize <= 1:
            return input_obj
        else:
            img = input_obj.get_data()
            blurred = cv2.GaussianBlur(img, (ksize, ksize), self.sigmaX, self.sigmaY)

            logger.debug("Added Gaussian Blur w/ Kernel Size=%d to Image" % (ksize,))
            return GenericImageEntity(blurred, input_obj.get_mask())


import logging
import os
from typing import Sequence, Any, Callable
import copy
import cloudpickle as pickle
import numpy as np
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.tensorboard import SummaryWriter
from torchtext.data.iterator import Iterator as TextDataIterator
from torchtext.data.iterator import BucketIterator
import torchtext
import torch.nn.utils.clip_grad as torch_clip_grad

from .datasets import CSVTextDataset
from .training_statistics import EpochStatistics, EpochTrainStatistics, EpochValidationStatistics
from .optimizer_interface import OptimizerInterface
from .default_optimizer import _running_eval_acc, _save_nandata, _validate_soft_to_hard_args
from .config import TorchTextOptimizerConfig
from .constants import VALID_OPTIMIZERS, MAX_EPOCHS

logger = logging.getLogger(__name__)


class TorchTextOptimizer(OptimizerInterface):
    """
    An optimizer for training and testing LSTM models. Currently in a prototype state.
    """

    def __init__(self, optimizer_cfg: TorchTextOptimizerConfig = None):
        """
        Initializes the optimizer with an TorchTextOptimizerConfig
        :param optimizer_cfg: the configuration used to initialize the TorchTextOptimizer
        """
        if optimizer_cfg is None:
            logger.info("Using default parameters to setup Optimizer!")
            self.optimizer_cfg = TorchTextOptimizerConfig()
        elif not isinstance(optimizer_cfg, TorchTextOptimizerConfig):
            msg = "optimizer_cfg must be of type TorchTextOptimizerConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.optimizer_cfg = optimizer_cfg

        # setup parameters for training here
        self.device = self.optimizer_cfg.training_cfg.device

        if not callable(self.optimizer_cfg.training_cfg.objective):
            self.loss_function_str = self.optimizer_cfg.training_cfg.objective.lower()
            if self.loss_function_str == "cross_entropy_loss":
                self.loss_function = nn.CrossEntropyLoss(**self.optimizer_cfg.training_cfg.objective_kwargs)
            elif self.loss_function_str == 'bcewithlogitsloss':
                self.loss_function = nn.BCEWithLogitsLoss(**self.optimizer_cfg.training_cfg.objective_kwargs)
            else:
                msg = self.loss_function_str + ": Unsupported objective function!"
                logger.error(msg)
                raise ValueError(msg)
        else:
            self.loss_function = self.optimizer_cfg.training_cfg.objective
        self.loss_function.to(self.device)

        self.lr = self.optimizer_cfg.training_cfg.lr
        # setup learning rate scheduler if desired
        self.lr_scheduler = None

        self.optimizer_str = self.optimizer_cfg.training_cfg.optim.lower()
        self.optimizer = None
        self.optim_kwargs = self.optimizer_cfg.training_cfg.optim_kwargs

        self.batch_size = self.optimizer_cfg.training_cfg.batch_size
        self.num_epochs = self.optimizer_cfg.training_cfg.epochs
        self.save_best_model = self.optimizer_cfg.training_cfg.save_best_model

        self.str_description = "{'batch_size':%d, 'num_epochs':%d, 'device':'%s', 'lr':%.5e, 'loss_function':'%s', " \
                               "'optimizer':'%s'}" % \
                               (self.batch_size, self.num_epochs, self.device.type, self.lr, self.loss_function_str,
                                self.optimizer_str)

        # setup configuration for logging and tensorboard
        self.num_batches_per_logmsg = self.optimizer_cfg.reporting_cfg.num_batches_per_logmsg
        self.num_epochs_per_metrics = self.optimizer_cfg.reporting_cfg.num_epochs_per_metrics
        self.num_batches_per_metrics = self.optimizer_cfg.reporting_cfg.num_batches_per_metrics

        # raise error if train/val split is not set properly for either saving best model
        # or for early stopping
        if self.optimizer_cfg.training_cfg.early_stopping or self.save_best_model:
            self.num_epochs_per_metrics = 1
            logger.warning("Overriding num_epochs_per_metrics due to early-stopping or saving-best-model!")

        if self.device.type == 'cpu' and self.num_batches_per_metrics:
            logger.warning('Training will be VERY SLOW on a CPU with num_batches_per_val_dataset_metrics set to a '
                           'value other than None.  If validation dataset metrics are still desired, '
                           'consider increasing this value to speed up training')

        tensorboard_output_dir = self.optimizer_cfg.reporting_cfg.tensorboard_output_dir
        self.tb_writer = None
        if tensorboard_output_dir:
            self.tb_writer = SummaryWriter(tensorboard_output_dir)

        optimizer_cfg_str = 'Optimizer[%s] Configured as: loss[%s], learning-rate[%.5e], batch-size[%d] ' \
                            'num-epochs[%d] Device[%s]' % \
                            (self.optimizer_str, str(self.loss_function), self.lr, self.batch_size, self.num_epochs,
                             self.device.type)
        reporting_cfg_str = 'Reporting Configured as: num_batches_per_log_message[%d] tensorboard_dir[%s]' % \
                            (self.num_batches_per_logmsg, tensorboard_output_dir)
        nbpm_print = self.num_batches_per_metrics if self.num_batches_per_metrics else -1
        metrics_capture_str = 'Metrics capturing configured as: num_epochs_per_metric[%d] ' \
                              'num_batches_per_epoch_per_metric[%d]' % \
                              (self.num_epochs_per_metrics, nbpm_print)
        logger.info(self.str_description)
        logger.info(optimizer_cfg_str)
        logger.info(reporting_cfg_str)
        logger.info(metrics_capture_str)

    def __str__(self):
        return self.str_description

    def __deepcopy__(self, memodict={}):
        optimizer_cfg_copy = copy.deepcopy(self.optimizer_cfg)
        # WARNING: this assumes that none of the derived attributes have been changed after construction!
        return TorchTextOptimizer(TorchTextOptimizerConfig(optimizer_cfg_copy.training_cfg,
                                                           optimizer_cfg_copy.reporting_cfg))

    def __eq__(self, other: Any):
        try:
            if self.optimizer_cfg == other.optimizer_cfg:
                # we still check the derived attributes to ensure that they remained the same after
                # after construction
                if self.device.type == other.device.type and self.loss_function_str == other.loss_function_str and \
                        self.lr == other.lr and self.optimizer_str == other.optimizer_str and \
                        self.batch_size == other.batch_size and self.num_epochs == other.num_epochs and \
                        self.str_description == other.str_description and \
                        self.num_batches_per_logmsg == other.num_batches_per_logmsg and \
                        self.num_epochs_per_metrics == other.num_epochs_per_metrics and \
                        self.num_batches_per_metrics == other.num_batches_per_metrics and \
                        self.tb_writer.log_dir == other.tb_writer.log_dir:
                    return True
            else:
                return False
        except AttributeError:
            return False

    def get_cfg_as_dict(self) -> dict:
        return self.optimizer_cfg.training_cfg.get_cfg_as_dict()

    def get_device_type(self) -> str:
        """
        :return: a string representing the device used to train the model
        """
        return self.device.type

    def save(self, fname: str) -> None:
        """
        Saves the configuration object used to construct the TorchTextOptimizer.
        NOTE: because the TorchTextOptimizer object itself is not persisted, but rather the
          TorchTextOptimizerConfig object, the state of the object does not persist!
        :param fname: the filename to save the TorchTextOptimizer's configuration.
        """
        self.optimizer_cfg.save(fname)

    @staticmethod
    def load(fname: str) -> OptimizerInterface:
        """
        Reconstructs an TorchTextOptimizer, by loading the configuration used to construct the original
        TorchTextOptimizer, and then creating a new TorchTextOptimizer object from the saved configuration
        :param fname: The filename of the saved TorchTextOptimizer
        :return: an TorchTextOptimizer object
        """
        with open(fname, 'rb') as f:
            loaded_optimzier_cfg = pickle.load(f)
        return TorchTextOptimizer(loaded_optimzier_cfg)

    def _eval_loss_function(self, y_hat: torch.Tensor, y_truth: torch.Tensor) -> torch.Tensor:
        """
        Wrapper for evaluating the loss function to abstract out any data casting we need to do
        :param y_hat: the predicted y-value
        :param y_truth: the actual y-value
        :return: the loss associated w/ the prediction and actual
        """
        if self.loss_function_str == "cross_entropy_loss":
            train_loss = self.loss_function(y_hat, y_truth.long())
        else:
            train_loss = self.loss_function(y_hat, y_truth)
        return train_loss

    @staticmethod
    def train_val_dataset_split(dataset: torchtext.data.Dataset, split_amt: float, val_data_transform: Callable,
                                val_label_transform: Callable) \
            -> (torchtext.data.Dataset, torchtext.data.Dataset):
        """
        Splits a torchtext dataset (of type: torchtext.data.Dataset) into train/test.
        NOTE: although this has the same functionality as default_optimizer.train_val_dataset_split, it works with a
         torchtext.data.Dataset object rather than torch.utils.data.Dataset.
        TODO:
          [ ] - specify random seed to torch splitter
        :param dataset: the dataset to be split
        :param split_amt: fraction specificing the validation dataset size relative to the whole.  1-split_amt will
                          be the size of the training dataset
        :param val_data_transform: (function: any -> any) how to transform the validation data to fit
                into the desired model and objective function
        :param val_label_transform: (function: any -> any) how to transform the validation labels
        :return: a tuple of the train and validation datasets
        """

        if split_amt < 0 or split_amt > 1:
            msg = "Dataset split amount must be between 0 and 1"
            logger.error(msg)
            raise ValueError(msg)
        if np.isclose(split_amt, 0.):
            train_dataset = dataset
            val_dataset = None
        else:
            train_dataset, val_dataset = dataset.split(1 - split_amt)
            val_dataset.data_transform = val_data_transform
            val_dataset.label_transform = val_label_transform
        return train_dataset, val_dataset

    @staticmethod
    def _eval_acc(data_loader, model, device=torch.device('cpu'),
                  soft_to_hard_fn: Callable = None,
                  soft_to_hard_fn_kwargs: dict = None,
                  loss_fn: Callable = None):
        """
        Evaluates a model against a dataset encompassed by a data loader, which has
        an underlying torchtext dataset.  The functionality is the same as default_optimizer._eval_acc,
        but used for torchtext.utils.Dataset rather than a torch.data.utils.Dataset
        """
        soft_to_hard_fn, soft_to_hard_fn_kwargs = _validate_soft_to_hard_args(soft_to_hard_fn,
                                                                              soft_to_hard_fn_kwargs)

        n_correct = None
        n_total = None
        model.eval()

        total_val_loss = 0.
        with torch.no_grad():
            for val_batch_idx, batch in enumerate(data_loader):
                if model.packed_padded_sequences:
                    text, text_lengths = batch.text
                    x = (text, text_lengths)
                    predictions = model(text, text_lengths).squeeze(1)
                else:
                    x = batch.text
                    predictions = model(batch.text).squeeze(1)

                if loss_fn is not None:
                    loss_tensor = loss_fn(predictions, batch.label)
                    batch_loss = loss_tensor.item()
                    total_val_loss += batch_loss

                running_acc, n_total, n_correct = _running_eval_acc(predictions, batch.label,
                                                                    n_total=n_total,
                                                                    n_correct=n_correct,
                                                                    soft_to_hard_fn=soft_to_hard_fn,
                                                                    soft_to_hard_fn_kwargs=soft_to_hard_fn_kwargs)

                if (loss_fn is not None and np.isnan(batch_loss)) or np.isnan(running_acc):
                    _save_nandata(x, predictions, batch.label, loss_tensor, batch_loss,
                                  running_acc, n_total, n_correct, model)

        total_val_loss /= float(len(data_loader))
        return running_acc, n_total, n_correct, total_val_loss

    def convert_dataset_to_dataiterator(self, dataset: CSVTextDataset, batch_size: int=None) -> TextDataIterator:
        # NOTE: we use the argument drop_last for the DataLoader (used for the CSVDataset), but no such argument
        # exists for the BucketIterator.  TODO: test whether this might become a problem.
        if not batch_size:
            batch_size_in = self.batch_size
        else:
            batch_size_in = batch_size
        return BucketIterator(dataset, batch_size_in, device=self.device, sort_within_batch=True)

    def train(self, net: torch.nn.Module, dataset: CSVTextDataset, progress_bar_disable: bool = False,
              torch_dataloader_kwargs: dict = None) -> (torch.nn.Module, Sequence[EpochStatistics], int):
        """
        Train the network.
        :param net: the model to train
        :param dataset: the dataset to train the network on
        :param progress_bar_disable: if True, disables the progress bar
        :param torch_dataloader_kwargs: additional arguments to pass to PyTorch's DataLoader class
        :return: the trained network, list of EpochStatistics objects, and the # of epochs on which teh net was trained
        """
        net = net.to(self.device)

        net.train()  # put network into training mode
        if self.optimizer_str == 'adam':
            self.optimizer = optim.Adam(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str == 'sgd':
            self.optimizer = optim.SGD(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str not in VALID_OPTIMIZERS:
            msg = self.optimizer_str + " is not a supported optimizer!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            msg = self.optimizer_str + " not yet implemented!"
            logger.error(msg)
            raise NotImplementedError(msg)
        if self.optimizer_cfg.training_cfg.lr_scheduler is not None:
            self.lr_scheduler = self.optimizer_cfg.training_cfg.lr_scheduler(self.optimizer,
                                                                             **self.optimizer_cfg.training_cfg.lr_scheduler_init_kwargs)


        # split into train & validation datasets, and setup data loaders according to their type
        train_dataset, val_dataset = TorchTextOptimizer.train_val_dataset_split(dataset,
                                                                                self.optimizer_cfg.training_cfg.train_val_split,
                                                                                self.optimizer_cfg.training_cfg.val_data_transform,
                                                                                self.optimizer_cfg.training_cfg.val_label_transform)
        train_loader = self.convert_dataset_to_dataiterator(train_dataset)
        val_loader = self.convert_dataset_to_dataiterator(val_dataset) if val_dataset else []

        # before training - we should transfer the embedding to the model weights if desired by the user
        if self.optimizer_cfg.copy_pretrained_embeddings:
            if hasattr(net, 'embedding'):
                pretrained_embeddings = dataset.text_field.vocab.vectors
                net.embedding.weight.data.copy_(pretrained_embeddings)
                # get the indices in the embedding which correspond to the UNK and the PAD characters
                UNK_IDX = dataset.text_field.vocab.stoi[dataset.text_field.unk_token]
                PAD_IDX = dataset.text_field.vocab.stoi[dataset.text_field.pad_token]
                # UNK_IDX and PAD_IDX are initialized to a N(0,1) distribution, per our arguments to the build_vocab function
                #  but we zero it out.
                #  Per: https://github.com/bentrevett/pytorch-sentiment-analysis/blob/master/2%20-%20Upgraded%20Sentiment%20Analysis.ipynb
                #  it is better to do this to train the model to konw that pad and unk are irrelevant in the classification task
                net.embedding.weight.data[UNK_IDX] = torch.zeros(net.embedding_dim)
                net.embedding.weight.data[PAD_IDX] = torch.zeros(net.embedding_dim)
            else:
                msg = "Cannot copy pretrained embeddings to network which doesn't have an attribute 'embedding'!"
                logger.error(msg)
                raise ValueError(msg)

        # use validation in training? provide as option?
        epoch_stats = []
        best_net = None
        best_validation_acc = -999
        best_val_loss = np.inf
        best_val_loss_epoch = -1

        num_epochs_to_monitor = 1
        if self.optimizer_cfg.training_cfg.early_stopping:
            num_epochs_to_monitor = self.optimizer_cfg.training_cfg.early_stopping.num_epochs

        epoch = 0
        done = False
        while not done:
            train_stats, validation_stats = self.train_epoch(net, train_loader, val_loader, epoch,
                                                             progress_bar_disable=progress_bar_disable)
            epoch_training_stats = EpochStatistics(epoch, train_stats, validation_stats)
            epoch_stats.append(epoch_training_stats)

            # TODO: save best model should use same criterion as early stopping (val-loss rather than val-acc)?
            if self.save_best_model:
                # use validation accuracy as the metric for deciding the best model
                if validation_stats.get_val_acc() >= best_validation_acc:
                    msg = "Updating best model with epoch:[%d] accuracy[%0.02f].  Previous best validation " \
                          "accuracy was: %0.02f" % (epoch, validation_stats.get_val_acc(), best_validation_acc)
                    logger.info(msg)
                    best_net = copy.deepcopy(net)
                    best_validation_acc = validation_stats.get_val_acc()

            # early stopping
            # record the val loss of the last batch in the epoch.  if N epochs after the best val_loss, we have not
            # improved the val-loss by atleast eps, we quit
            if self.optimizer_cfg.training_cfg.early_stopping:
                # EarlyStoppingConfig validates that eps > 0 as well ..
                if validation_stats.get_val_loss() < (
                        best_val_loss - np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps)):
                    best_val_loss = validation_stats.get_val_loss()
                    best_val_loss_epoch = epoch
                    best_net = copy.deepcopy(net)
                    logger.info('EarlyStopping - NewBest >> best_val_loss:%0.04f best_val_loss_epoch:%d' %
                                (best_val_loss, best_val_loss_epoch))
                elif epoch >= (best_val_loss_epoch + num_epochs_to_monitor):
                    epoch += 1  # we do this b/c of the break to keep the accounting of epoch # returned to
                    # the user to be one based
                    msg = "Exiting training loop in epoch: %d - due to early stopping criterion being met!" \
                          % (epoch,)
                    logger.warning(msg)
                    done = True

            epoch += 1
            if self.optimizer_cfg.training_cfg.early_stopping:
                # in case something goes wrong, we exit after training a long time ...
                if epoch >= MAX_EPOCHS:
                    done = True
            else:
                if epoch >= self.num_epochs:
                    done = True

        if self.save_best_model or self.optimizer_cfg.training_cfg.early_stopping:
            return best_net, epoch_stats, epoch, best_val_loss_epoch
        else:
            return net, epoch_stats, epoch, best_val_loss_epoch

    def train_epoch(self, model: nn.Module, train_loader: TextDataIterator, val_loader: TextDataIterator,
                    epoch_num: int, progress_bar_disable: bool = False):
        """
        Runs one epoch of training on the specified model

        :param model: the model to train for one epoch
        :param train_loader: a DataLoader object pointing to the training dataset
        :param val_loader: a DataLoader object pointing to the validation dataset
        :param epoch_num: the epoch number that is being trained
        :param progress_bar_disable: if True, disables the progress bar
        :return: a list of statistics for batches where statistics were computed
        """

        pid = os.getpid()
        train_dataset_len = len(train_loader.dataset)
        loop = tqdm(train_loader, disable=progress_bar_disable)

        train_n_correct, train_n_total = None, None
        val_n_correct, val_n_total = None, None
        sum_batchmean_train_loss = 0
        running_train_acc = 0
        num_batches = len(train_loader)
        # put network into training mode
        model.train()
        for batch_idx, batch in enumerate(loop):
            # zero out previous gradient computations
            self.optimizer.zero_grad()

            # get predictions based on input & weights learned so far
            if model.packed_padded_sequences:
                text, text_lengths = batch.text
                x = (text, text_lengths)
                predictions = model(text, text_lengths).squeeze(1)
            else:
                x = batch.text
                predictions = model(batch.text).squeeze(1)

            # compute metrics
            batch_train_loss = self._eval_loss_function(predictions, batch.label)
            sum_batchmean_train_loss += batch_train_loss.item()
            running_train_acc, train_n_total, train_n_correct = \
                _running_eval_acc(predictions, batch.label, n_total=train_n_total, n_correct=train_n_correct,
                                  soft_to_hard_fn=self.optimizer_cfg.training_cfg.soft_to_hard_fn,
                                  soft_to_hard_fn_kwargs=self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs)

            if np.isnan(sum_batchmean_train_loss) or np.isnan(running_train_acc):
                _save_nandata(x, predictions, batch.label, batch_train_loss, sum_batchmean_train_loss, running_train_acc,
                              train_n_total, train_n_correct, model)

            # compute gradient
            batch_train_loss.backward()

            # perform gradient clipping if configured
            if self.optimizer_cfg.training_cfg.clip_grad:
                if self.optimizer_cfg.training_cfg.clip_type == 'norm':
                    # clip_grad_norm_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_norm_(model.parameters(), self.optimizer_cfg.training_cfg.clip_val,
                                                    **self.optimizer_cfg.training_cfg.clip_kwargs)
                elif self.optimizer_cfg.training_cfg.clip_type == 'val':
                    # clip_grad_val_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_value_(model.parameters(), self.optimizer_cfg.training_cfg.clip_val)
                else:
                    msg = "Unknown clipping type for gradient clipping!"
                    logger.error(msg)
                    raise ValueError(msg)

            self.optimizer.step()

            loop.set_description('Epoch {}/{}'.format(epoch_num + 1, self.num_epochs))
            loop.set_postfix(avg_train_loss=batch_train_loss.item())

            # report batch statistics to tensorboard
            if self.tb_writer:
                try:
                    batch_num = int(epoch_num * num_batches + batch_idx)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-train_loss',
                                              batch_train_loss.item(), global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-running_train_acc',
                                              running_train_acc, global_step=batch_num)
                except:
                    # TODO: catch specific exceptions!
                    pass

            if batch_idx % self.num_batches_per_logmsg == 0:
                logger.info('{}\tTrain Epoch: {} [{}/{} ({:.0f}%)]\tTrainLoss: {:.6f}\tTrainAcc: {:.6f}'.format(
                    pid, epoch_num, batch_idx * len(batch), train_dataset_len,
                                    100. * batch_idx / num_batches, batch_train_loss.item(), running_train_acc))
        train_stats = EpochTrainStatistics(running_train_acc, sum_batchmean_train_loss / float(num_batches))

        # if we have validation data, we compute on the validation dataset
        validation_stats = None
        num_val_batches = len(val_loader)
        if num_val_batches > 0:
            logger.info('Running validation')
            val_acc, _, _, val_loss = TorchTextOptimizer._eval_acc(val_loader, model, device=self.device,
                                                                   soft_to_hard_fn=self.optimizer_cfg.training_cfg.soft_to_hard_fn,
                                                                   soft_to_hard_fn_kwargs=self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs,
                                                                   loss_fn=self._eval_loss_function)
            validation_stats = EpochValidationStatistics(val_acc, val_loss, None, None)

            logger.info('{}\tTrain Epoch: {} \tValLoss: {:.6f}\tValAcc: {:.6f}'.format(
                pid, epoch_num, val_loss, val_acc))

            if self.tb_writer:
                try:
                    batch_num = int((epoch_num + 1) * num_batches)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-validation_loss', val_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-validation_acc', val_acc, global_step=batch_num)
                except:
                    # TODO: catch specific exceptions!
                    pass

        # update the lr-scheduler if necessary
        if self.lr_scheduler is not None:
            if self.optimizer_cfg.training_cfg.lr_scheduler_call_arg is None:
                self.lr_scheduler.step()
            elif self.optimizer_cfg.training_cfg.lr_scheduler_call_arg.lower() == 'val_acc':
                if num_val_batches > 0:  # this check ensures that this variable is defined
                    self.lr_scheduler.step(val_acc)
                else:
                    msg = "val_acc not defined b/c validation dataset is not defined! Ignoring LR step!"
                    logger.warning(msg)
            elif self.optimizer_cfg.training_cfg.lr_scheduler_call_arg.lower() == 'val_loss':
                if num_val_batches > 0:
                    self.lr_scheduler.step(val_loss)
                else:
                    msg = "val_loss not defined b/c validation dataset is not defined! Ignoring LR step!"
                    logger.warning(msg)
            else:
                msg = "Unknown mode for calling lr_scheduler!"
                logger.error(msg)
                raise ValueError(msg)

        return train_stats, validation_stats

    def test(self, model: nn.Module, clean_data: CSVTextDataset, triggered_data: CSVTextDataset,
             clean_test_triggered_labels_data: CSVTextDataset, progress_bar_disable: bool = False,
             torch_dataloader_kwargs: dict = None) -> dict:
        """
        Test the trained network
        :param model: the trained module to run the test data through
        :param clean_data: the clean Dataset
        :param triggered_data: the triggered Dataset, if None, not computed
        :param clean_test_triggered_labels_data: triggered part of the training dataset but with correct labels; see
            DataManger.load_data for more information.
        :param progress_bar_disable: if True, disables the progress bar
        :param torch_dataloader_kwargs: additional arguments to pass to PyTorch's DataLoader class
        :return: a dictionary of the statistics on the clean and triggered data (if applicable)
        """
        test_data_statistics = {}
        model.eval()

        # setup for test data batch-size = 1, so that we don't drop last batch if it does not fit fully into a batch
        # see: https://pytorch.org/docs/stable/data.html#data-loading-order-and-sampler
        data_loader = self.convert_dataset_to_dataiterator(clean_data, 1)
        test_acc, test_n_total, _, _ = TorchTextOptimizer._eval_acc(data_loader, model, device=self.device,
                                                                    soft_to_hard_fn=self.optimizer_cfg.training_cfg.soft_to_hard_fn,
                                                                    soft_to_hard_fn_kwargs=self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs,
                                                                    loss_fn=None)
        test_data_statistics['clean_accuracy'] = test_acc
        test_data_statistics['clean_n_total'] = test_n_total
        logger.info("Accuracy on clean test data: %0.02f" %
                    (test_data_statistics['clean_accuracy'],))

        if triggered_data is None:
            return test_data_statistics

        # setup for test data batch-size = 1, so that we don't drop last batch if it does not fit fully into a batch
        # see: https://pytorch.org/docs/stable/data.html#data-loading-order-and-sampler
        data_loader = self.convert_dataset_to_dataiterator(triggered_data, 1)
        test_acc, test_n_total, _, _ = TorchTextOptimizer._eval_acc(data_loader, model, device=self.device,
                                                                    soft_to_hard_fn=self.optimizer_cfg.training_cfg.soft_to_hard_fn,
                                                                    soft_to_hard_fn_kwargs=self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs,
                                                                    loss_fn=None)
        test_data_statistics['triggered_accuracy'] = test_acc
        test_data_statistics['triggered_n_total'] = test_n_total
        logger.info("Accuracy on triggered test data: %0.02f" %
                    (test_data_statistics['triggered_accuracy'],))

        # Test the classification accuracy on clean data for labels which have corresponding triggered examples.
        # For example, if an MNIST dataset was created with triggered examples only for labels 4 and 5,
        # then this dataset is the subset of data with labels 4 and 5 that don't have the triggers.
        data_loader = self.convert_dataset_to_dataiterator(clean_test_triggered_labels_data, 1)
        test_acc, test_n_total, _, _ = TorchTextOptimizer._eval_acc(data_loader, model, device=self.device,
                                                                    soft_to_hard_fn=self.optimizer_cfg.training_cfg.soft_to_hard_fn,
                                                                    soft_to_hard_fn_kwargs=self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs,
                                                                    loss_fn=None)
        test_data_statistics['clean_test_triggered_label_accuracy'] = test_acc
        test_data_statistics['clean_test_triggered_label_n_total'] = test_n_total
        logger.info("Accuracy on clean-data-triggered-labels: %0.02f for n=%s" %
                    (test_data_statistics['clean_test_triggered_label_accuracy'], str(test_n_total)))

        return test_data_statistics
